"""Flow, pressure changes, and production from hydraulically fractured reservoirs."""
from __future__ import annotations

from bluebonnet.flow.flowproperties import (
    FlowProperties,
    FlowPropertiesMultiPhase,
    FlowPropertiesOnePhase,
    FlowPropertiesTwoPhase,
    RelPermParams,
    relative_permeabilities,
    relative_permeabilities_twophase,
    rescale_pseudopressure,
)
from bluebonnet.flow.reservoir import (
    IdealReservoir,
    MultiPhaseReservoir,
    SinglePhaseReservoir,
    TwoPhaseReservoir,
)

__all__ = [
    "FlowProperties",
    "FlowPropertiesMultiPhase",
    "FlowPropertiesOnePhase",
    "FlowPropertiesTwoPhase",
    "RelPermParams",
    "relative_permeabilities",
    "IdealReservoir",
    "MultiPhaseReservoir",
    "SinglePhaseReservoir",
    "TwoPhaseReservoir",
    "relative_permeabilities_twophase",
    "rescale_pseudopressure",
]



"""Scaling solutions for production analysis from unconventional oil and gas wells."""

from __future__ import annotations

from importlib.metadata import version

__version__ = version("bluebonnet")

import bluebonnet.plotting



"""Define a suite a tests for the gas module."""
from __future__ import annotations

import numpy as np
import pytest
from bluebonnet.fluids import gas

TEMPERATURE_STANDARD = 60.0
PRESSURE_STANDARD = 14.70


def test_make_nonhydrocarbon_properties():
    """Test creation of non-hydrocarbon molecular properties."""
    nitrogen = 0.03
    hydrogen_sulfide = 0.012
    co2 = 0.018
    dtypes = [
        ("name", "U20"),
        ("fraction", "f8"),
        ("molecular weight", "f8"),
        ("critical temperature", "f8"),
        ("critical pressure", "f8"),
    ]
    correct_array = np.array(
        [
            ("Nitrogen", nitrogen, 28.01, 226.98, 492.26),
            ("Hydrogen sulfide", hydrogen_sulfide, 34.08, 672.35, 1299.97),
            ("CO2", co2, 44.01, 547.54, 1070.67),
        ],
        dtype=dtypes,
    )

    test_array = gas.make_nonhydrocarbon_properties(nitrogen, hydrogen_sulfide, co2)
    for col in (
        "fraction",
        "molecular weight",
        "critical temperature",
        "critical pressure",
    ):
        np.testing.assert_allclose(test_array[col], correct_array[col])

    helium = 0.01
    others = [("Helium", helium, 4.0, 226.8, 493.1)]

    gas.make_nonhydrocarbon_properties(nitrogen, hydrogen_sulfide, co2, *others)


def test_pseudocritical_points_Sutton(gas_properties):
    """Test Sutton pseudocritical pressure and temperature."""
    fluid = gas_properties.fluid
    specific_gravity = gas_properties.specific_gravity
    temperature_pc_true = pytest.approx(gas_properties.temperature_pc)
    pressure_pc_true = pytest.approx(gas_properties.pressure_pc)
    if fluid == "dry gas":
        fractions = (0.03, 0.012, 0.018)
    elif fluid == "wet gas":
        fractions = (0.05, 0.01, 0.04)
    else:
        raise NotImplementedError
    non_hydrocarbon_properties = gas.make_nonhydrocarbon_properties(*fractions)
    temperature_pc, pressure_pc = gas.pseudocritical_point_Sutton(
        specific_gravity, non_hydrocarbon_properties, fluid
    )
    assert pressure_pc == pressure_pc_true
    assert temperature_pc == temperature_pc_true


def test_zfactor_DAK(gas_properties):
    """Test Dranchuk's z-factor calculation."""
    (
        temperature,
        pressure,
        temperature_pc,
        pressure_pc,
        _,
        fluid,
    ) = gas_properties
    if fluid == "dry gas":
        real_z = pytest.approx(0.996901, rel=1e-3)
    else:
        real_z = pytest.approx(1.066513, rel=1e-3)
    z_factor_DAK = gas.z_factor_DAK(temperature, pressure, temperature_pc, pressure_pc)
    assert z_factor_DAK == real_z


def test_b_factor_DAK(gas_properties):
    """Test Dranchuk's b_g (depends on z-factor)."""
    (
        temperature,
        pressure,
        temperature_pc,
        pressure_pc,
        _,
        fluid,
    ) = gas_properties
    if fluid == "dry gas":
        real_b_g = pytest.approx(0.04317415921420302, rel=1e-3)
    else:
        real_b_g = pytest.approx(8.139289641971048e-4, rel=1e-3)
    b_factor = gas.b_factor_DAK(
        temperature,
        pressure,
        temperature_pc,
        pressure_pc,
        TEMPERATURE_STANDARD,
        PRESSURE_STANDARD,
    )
    assert b_factor == real_b_g


def test_density_DAK(gas_properties):
    """Test Dranchuk's density calculation (depends on z-factor)."""
    (
        temperature,
        pressure,
        temperature_pc,
        pressure_pc,
        gas_specific_gravity,
        fluid,
    ) = gas_properties
    if fluid == "dry gas":
        real_density = pytest.approx(0.204702, rel=1e-3)
    else:
        real_density = pytest.approx(12.79783, rel=1e-3)
    density = gas.density_DAK(
        temperature, pressure, temperature_pc, pressure_pc, gas_specific_gravity
    )
    assert density == real_density


def test_compressibility_DAK(gas_properties):
    """Test Dranchuk's compressibility (depends on z-factor)."""
    (
        temperature,
        pressure,
        temperature_pc,
        pressure_pc,
        gas_specific_gravity,
        fluid,
    ) = gas_properties
    if fluid == "dry gas":
        real_compressibility = pytest.approx(0.01002548578259275, rel=1e-3)
    else:
        real_compressibility = pytest.approx(1.4644387216173005e-4, rel=1e-3)
    compressibility = gas.compressibility_DAK(
        temperature, pressure, temperature_pc, pressure_pc
    )
    assert compressibility == real_compressibility


def test_viscosity_Sutton(gas_properties):
    """Test Sutton's viscosity (depends on density, which depends on z-factor)."""
    (
        temperature,
        pressure,
        temperature_pc,
        pressure_pc,
        gas_specific_gravity,
        fluid,
    ) = gas_properties
    if fluid == "dry gas":
        real_viscosity = pytest.approx(1.6528333290904862e-2, rel=1e-3)
    else:
        real_viscosity = pytest.approx(2.3553396039641815e-2, rel=1e-3)
    viscosity = gas.viscosity_Sutton(
        temperature, pressure, temperature_pc, pressure_pc, gas_specific_gravity
    )
    assert viscosity == real_viscosity


def test_pseudopressure_Hussainy(gas_properties):
    """Test Al Hussainy's pseudopressure calculation (depends on viscosity, density, z-factor)."""
    (
        temperature,
        pressure,
        temperature_pc,
        pressure_pc,
        gas_specific_gravity,
        fluid,
    ) = gas_properties
    if fluid == "dry gas":
        real_m = pytest.approx(593363, rel=1e-3)
    else:
        real_m = pytest.approx(1280589747, rel=1e-3)
    m = gas.pseudopressure_Hussainy(
        temperature,
        pressure,
        temperature_pc,
        pressure_pc,
        gas_specific_gravity,
        PRESSURE_STANDARD,
    )
    assert m == real_m



"""Fit and forecast production from hydrofractured reservoirs."""
from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass

import numpy as np
import numpy.typing as npt
from numpy import ndarray
from scipy.optimize import curve_fit


@dataclass(frozen=True)
class Bounds:
    """Set the upper and lower limits for the fitting curve.

    Parameters
    ----------
    M: tuple of floats
        (min, max) for resource in place
    tau: tuple of floats
        (min, max) for time-to-BDF
    """

    M: tuple[float, float]
    """Minimum and maximum resource in place"""
    tau: tuple[float, float]
    """Minimum and maximum time-to-boundary dominated flow"""

    def __post_init__(self):
        """Validate bounds."""
        if len(self.M) != 2:
            msg = f"M must be two elements, not {len(self.M)} elements"
            raise ValueError(msg)
        if len(self.tau) != 2:
            msg = f"tau must be two elements, not {len(self.tau)} elements"
            raise ValueError(msg)
        if self.M[0] >= self.M[1]:
            msg = f"{self.M[0]=} must be greater than {self.M[1]=}"
            raise ValueError(msg)
        if self.tau[0] >= self.tau[1]:
            msg = f"{self.tau[0]=} must be greater than {self.tau[1]=}"
            raise ValueError(msg)

    def fit_bounds(self):
        """Return bounds for forecaster instance fits."""
        return ((self.M[0], self.tau[0]), (self.M[1], self.tau[1]))


_default_bounds = Bounds(M=(0, np.inf), tau=(1e-10, np.inf))


@dataclass
class ForecasterOnePhase:
    """Forecaster for production decline models.

    Parameters
    ----------
    rf_curve: Callable
        the recovery factor over scaled time, as a callable
        (interpolators are best). So :code:`rf_curve(time_scaled) -> recovery_factor`
    bounds : Bounds
        the minimum and maximum values that the rf_curve accepts
        (watch the limits for the rf_curve function)

    """

    rf_curve: Callable
    """function or interpolator with recovery factor over scaled time"""
    bounds: Bounds = _default_bounds
    """Limits on max and min resource in place and time-to-BDF"""

    def forecast_cum(
        self,
        time_on_production: npt.NDArray[np.float64],
        M: float | None = None,
        tau: float | None = None,
    ):
        """Forecast cumulative production.

        Parameters
        ----------
        time_on_production: ndarray
            time since the well started producing
            (days or years, ideally, since months are uneven)
        M: float
            the resource in place (Mstb, Mscf, or other standard units)
        tau: float
            the time until BDF/depletion flow
            (days or years, same units as time on prod)
        """
        if M is None:
            M = self.M_
        if tau is None:
            tau = self.tau_
        return _forecast_cum_onephase(self.rf_curve, time_on_production, M, tau)

    def fit(
        self,
        time_on_production: ndarray,
        cum_production: ndarray,
        tau: float | None = None,
    ):
        """Fit well production to the physics-based scaling model.

        Parameters
        ----------
        time_on_production: ndarray
            time since the well started producing
            (days or years, ideally, since months are uneven)
        cum_production: ndarray
            the cumulative production over time
            (Mstb, Mscf, or other standard units)
        tau: [Optional float]
            the time-to-depletion (BDF), using the same units as time on prod.
            If not provided, will find best tau.
        """
        if tau is None:
            p0 = (cum_production[-1] * 2, time_on_production[-1] * 5)
            bounds = self.bounds.fit_bounds()
            forecast = lambda top, M, tau: _forecast_cum_onephase(  # noqa: E731
                self.rf_curve, top, M, tau
            )
        else:
            p0 = (cum_production[-1] * 2,)
            bounds = self.bounds.M
            forecast = lambda top, M: _forecast_cum_onephase(  # noqa: BLK,E731
                self.rf_curve, top, M, tau
            )

        fit, covariance = curve_fit(
            forecast,
            time_on_production,
            cum_production,
            p0,
            bounds=bounds,
        )
        self.time_on_production = time_on_production
        self.cum_production = cum_production
        if tau is None:
            self.M_, self.tau_ = fit
        else:
            self.M_ = fit[0]
            self.tau_ = tau


def _forecast_cum_onephase(
    rf_curve: Callable, time_on_production: ndarray, M: float, tau: float
) -> ndarray:
    time_scaled = time_on_production / tau
    rf = M * rf_curve(time_scaled)
    return rf



"""Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

from __future__ import annotations

# -- Project information -----------------------------------------------------
project = "bluebonnet"
copyright = "2021-2023, Frank Male"
author = "Frank Male"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_nb",
    "autoapi.extension",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]
autoapi_dirs = ["../src"]
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# myst_nb configuration
nb_execution_mode = "off"
nb_execution_timeout = 60
nb_output_stderr = "remove"
myst_enable_extensions = [
    "amsmath",
    "dollarmath",
]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"



"""PVT and viscosity for water from the correlations provided by McCain."""
from __future__ import annotations

import numpy as np
from numpy.typing import NDArray


def b_water_McCain(
    temperature: float, pressure: float | NDArray[np.float64]
) -> float | NDArray[np.float64]:
    """Calculate the b-factor for water.

    Parameters
    ----------
    temperature : float
        water temperature in Fahrenheit
    pressure: np.ndarray
        water pressure in psia

    Returns
    -------
    b_w : float
        b-factor (reservoir bbl / standard bbl)

    Examples
    --------
    >>> b_water_McCain(400, 3000)

    """
    dV_dp = (
        -1.95301e-9 * pressure * temperature
        - 1.72834e-13 * pressure**2 * temperature
        - 3.58922e-7 * pressure
        - 2.25341e-10 * pressure**2
    )
    dV_dt = -1.0001e-2 + 1.33391e-4 * temperature + 5.50654e-7 * temperature**2
    return (1 + dV_dp) * (1 + dV_dt)


def b_water_McCain_dp(
    temperature: float, pressure: float | NDArray[np.float64]
) -> float | NDArray[np.float64]:
    """Calculate the derivative of the b-factor for water with respect to pressure.

    Parameters
    ----------
    temperature : float
        water temperature in Fahrenheit
    pressure: float | NDArray
        water pressure in psia

    Returns
    -------
    b_w_dp : float
        derivative of b-factor (reservoir bbl / standard bbl) / psi

    Examples
    --------
    >>> b_water_McCain_dp(400, 3000)

    """
    d2V_dp2 = (
        -1.95301e-9 * temperature
        - 2 * 1.72834e-13 * pressure * temperature
        - 3.58922e-7
        - 2 * 2.25341e-10 * pressure
    )
    dV_dt = -1.0001e-2 + 1.33391e-4 * temperature + 5.50654e-7 * temperature**2
    return d2V_dp2 * (1 + dV_dt)


def compressibility_water_McCain(
    temperature: float, pressure: float | NDArray[np.float64], salinity: float
) -> float | NDArray[np.float64]:
    """Calculate the compressibility for water.

    Parameters
    ----------
    temperature : float
        water temperature in Fahrenheit
    pressure: float | NDArray
        water pressure in psia
    salinity: float
        salinity in weight percent total dissolved solids

    Returns
    -------
    c_w : float
        compressibility of water in 1/psi

    Examples
    --------
    >>> density_water_McCain(400, 3000, 15)

    """
    c_w = 1 / (7.033 * pressure + 0.5415 * salinity - 537 * temperature + 403300.0)
    return c_w


def density_water_McCain(
    temperature: float, pressure: float | NDArray[np.float64], salinity: float
) -> float | NDArray[np.float64]:
    r"""Calculate the density for water.

    Parameters
    ----------
    temperature : float
        water temperature in Fahrenheit
    pressure: float | NDArray
        water pressure in psia
    salinity: float
        salinity in weight percent total dissolved solids

    Returns
    -------
    rho_w : float
        density in lb-mass / cu ft, :math:`\rho_w`

    Examples
    --------
    >>> density_water_McCain(400, 3000, 15)

    """
    b_water = b_water_McCain(temperature, pressure)
    density_stp = 62.368 + 0.438603 * salinity + 1.60074e-3 * salinity**2
    return density_stp / b_water


def viscosity_water_McCain(
    temperature: float, pressure: float | NDArray[np.float64], salinity: float
) -> float | NDArray[np.float64]:
    r"""Calculate the viscosity for water, Using McCain (1991).

    Parameters
    ----------
    temperature : float
        water temperature in Fahrenheit
    pressure: float
        water pressure in psia
    salinity: float
        salinity in weight percent total dissolved solids

    Returns
    -------
    mu_w : float
        viscosity in centipoise, :math:`\mu_w`

    Examples
    --------
    >>> viscosity_water_McCain(400, 3000, 15)
    0.2627774655403418
    """
    A = (
        109.574
        - 8.40564 * salinity
        + 0.313314 * salinity**2
        + 8.72213e-3 * salinity**3
    )
    B = (
        1.12166
        - 2.63951e-2 * salinity
        + 6.79461e-4 * salinity**2
        + 5.47119e-5 * salinity**3
        - 1.55586e-6 * salinity**4
    )
    mu_water = (
        A
        * temperature**-B
        * (0.9994 + 4.0295e-5 * pressure + 3.1062e-9 * pressure**2)
    )
    return mu_water



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
from bluebonnet.flow import FlowProperties
from bluebonnet.plotting import (
    SinglePhaseReservoir,
    plot_pseudopressure,
    plot_recovery_factor,
    plot_recovery_rate,
)

nx = 30
nt = 1000
pf = 100.0
pi = 2e3
columns_renamer_gas = {
    "P": "pressure",
    "Z-Factor": "z-factor",
    "Cg": "compressibility",
    "Viscosity": "viscosity",
    "Density": "density",
}
pvt_gas = pd.read_csv("tests/data/pvt_gas.csv").rename(columns=columns_renamer_gas)
fluid = FlowProperties(pvt_gas, pi)


@pytest.fixture()
def reservoir():
    reservoir = SinglePhaseReservoir(
        nx, pressure_fracface=pf, pressure_initial=pi, fluid=fluid
    )
    t_end = 11
    time_scaled = np.linspace(0, np.sqrt(t_end), nt) ** 2
    reservoir.simulate(time_scaled)
    return reservoir


@pytest.mark.mpl_image_compare()
@pytest.mark.parametrize("rescale", [True, False])
def test_plot_pseudopressure(reservoir, rescale):
    fig, ax = plt.subplots()
    plot_pseudopressure(reservoir, every=50, ax=ax, rescale=rescale)
    return fig


@pytest.mark.mpl_image_compare()
@pytest.mark.parametrize("change_ticks", [True, False])
def test_plot_recovery_rate(reservoir, change_ticks):
    fig, ax = plt.subplots()
    plot_recovery_rate(reservoir, ax, change_ticks=change_ticks)
    return fig


@pytest.mark.mpl_image_compare()
@pytest.mark.parametrize("change_ticks", [True, False])
def test_plot_recovery_factor(reservoir, change_ticks):
    fig, ax = plt.subplots()
    plot_recovery_factor(reservoir, ax, change_ticks=change_ticks)
    return fig



"""Fixtures for testing fluids modules."""

from __future__ import annotations

from typing import Any, NamedTuple

import pytest


class WaterParams(NamedTuple):
    temperature: float
    pressure: float
    salinity: float
    gas_saturated: float
    gas_properties: Any


class OilParams(NamedTuple):
    temperature: float
    pressure: float
    api_gravity: float
    gas_specific_gravity: float
    solution_gor_initial: float
    fluid: str


class GasParams(NamedTuple):
    temperature: float
    pressure: float
    temperature_pc: float
    pressure_pc: float
    specific_gravity: float
    fluid: str


@pytest.fixture(params=[True, False])
def water_properties(request):
    temperature = 200
    pressure = 4000
    salinity = 15
    gas_saturated = request.param
    gas_properties = None
    return WaterParams(temperature, pressure, salinity, gas_saturated, gas_properties)


@pytest.fixture(params=[(3000.0, "black oil"), (2000.0, "black oil")])
def oil_properties(request):
    pressure, fluid = request.param
    if fluid == "black oil":
        out = OilParams(
            temperature=200.0,
            pressure=pressure,
            api_gravity=35.0,
            gas_specific_gravity=0.8,
            solution_gor_initial=650.0,
            fluid="black oil",
        )
    return out


@pytest.fixture(params=["dry gas", "wet gas"])
def gas_properties(request):
    if request.param == "dry gas":
        out = GasParams(
            temperature=400,
            pressure=100,
            temperature_pc=-102.21827232417752,
            pressure_pc=648.510797253794,
            specific_gravity=0.65,
            fluid=request.param,
        )
    else:
        out = GasParams(
            temperature=300,
            pressure=5014.7,
            temperature_pc=-80.95111110103215,
            pressure_pc=656.7949325583305,
            specific_gravity=0.7661054354004884,
            fluid=request.param,
        )
    return out



"""Define a suite a tests for the flowproperties module."""
from __future__ import annotations

from copy import copy

import numpy as np
import pandas as pd
import pytest
from bluebonnet.flow.flowproperties import (
    FlowPropertiesTwoPhase,
    RelPermParams,
    relative_permeabilities,
    relative_permeabilities_twophase,
    rescale_pseudopressure,
)

pr = 8_000.0
Sw = 0.1
columns_renamer_gas = {
    "P": "pressure",
    "Z-Factor": "z-factor",
    "Cg": "compressibility",
    "Viscosity": "viscosity",
    "Density": "density",
}
columns_renamer_oil = {
    "P": "pressure",
    "Z-Factor": "z-factor",
    "Co": "compressibility",
    "Oil_Viscosity": "viscosity",
    "Oil_Density": "density",
}
pvt_gas = pd.read_csv("tests/data/pvt_gas.csv").rename(columns=columns_renamer_gas)
pvt_oil = pd.read_csv("tests/data/pvt_oil.csv").rename(columns=columns_renamer_oil)


@pytest.fixture()
def relperm_params():
    return RelPermParams(
        n_o=1,
        n_g=1,
        n_w=1,
        S_or=0,
        S_gc=0,
        S_wc=0.1,
        k_ro_max=1,
        k_rw_max=1,
        k_rg_max=1,
    )


@pytest.fixture()
def saturations_test():
    return pd.DataFrame(
        {
            "So": np.linspace(0, 1 - Sw, 50),
            "Sw": np.full(50, Sw),
            "Sg": np.linspace(1 - Sw, 0, 50),
        }
    )


@pytest.fixture()
def df_pvt():
    # get pvt tables
    pvt_oil = pd.read_csv("tests/data/pvt_oil.csv")
    pvt_water = pd.read_csv("tests/data/pvt_water.csv").rename(
        columns={"T": "temperature", "P": "pressure", "Viscosity": "mu_w"}
    )
    rename_cols = {
        "T": "temperature",
        "P": "pressure",
        "Oil_Viscosity": "mu_o",
        "Gas_Viscosity": "mu_g",
        "Rso": "Rs",
    }
    df_pvt = (
        pvt_water.drop(columns=["temperature"])
        .merge(
            pvt_oil.rename(columns=rename_cols),
            on="pressure",
        )
        .assign(Rv=0)
    )
    # calculate So, Sg assuming no mobile water
    df_pvt["So"] = (1 - Sw) / (
        (df_pvt["Rs"].max() - df_pvt["Rs"]) * df_pvt["Bg"] / df_pvt["Bo"] / 5.61458 + 1
    )
    return df_pvt


# # TODO: Figure out how to test FlowProperties
# @pytest.mark.parametrize("pvt_table", [pvt_gas, pvt_oil])
# def test_flow_properties(pvt_table):
#     alpha = 1 / (pvt_table["viscosity"] * pvt_table["density"])
#     fluid = FlowProperties(pvt_table, pr)
#     assert np.allclose(alpha, fluid.pvt_props["alpha"])


def test_relative_perms(relperm_params):
    """Relative permeability testing."""
    df_kr = relative_permeabilities_twophase(relperm_params)
    for fluid in ["o", "g"]:
        assert (
            df_kr.iloc[df_kr[f"S{fluid}"].argmax()][f"kr{fluid}"]
            > df_kr.iloc[df_kr[f"S{fluid}"].argmin()][f"kr{fluid}"]
        ), "relative permeability should be higher at higher saturation"
    assert df_kr["krw"].max() < 1e-3, "Water should be immobile below S_wr"


def test_relperm_bounds_errors(relperm_params, saturations_test):
    saturations = copy(saturations_test)
    saturations += 1
    with pytest.raises(ValueError, match="Saturations must sum to 1"):
        relative_permeabilities(saturations.to_records(index=False), relperm_params)
    # test parameter bounds
    relperm_wrong = relperm_params._replace(n_o=8)
    with pytest.raises(ValueError, match="Exponents .* less than"):
        relative_permeabilities(saturations_test.to_records(index=False), relperm_wrong)
    relperm_wrong = relperm_params._replace(n_w=0)
    with pytest.raises(ValueError, match="Exponents .* at least"):
        relative_permeabilities(saturations_test.to_records(index=False), relperm_wrong)
    relperm_wrong = relperm_params._replace(S_gc=-1)
    with pytest.raises(ValueError, match="saturation.* at least"):
        relative_permeabilities(saturations_test.to_records(index=False), relperm_wrong)
    relperm_wrong = relperm_params._replace(S_or=1.1)
    with pytest.raises(ValueError, match="saturation.* less than"):
        relative_permeabilities(saturations_test.to_records(index=False), relperm_wrong)
    relperm_wrong = relperm_params._replace(k_ro_max=1.1)
    with pytest.raises(ValueError, match="relative permeability.* less than"):
        relative_permeabilities(saturations_test.to_records(index=False), relperm_wrong)


def test_relative_perm_twophase(relperm_params):
    with pytest.raises(ValueError, match="Water saturation is above residual"):
        relative_permeabilities_twophase(relperm_params, Sw=0.8)
    df_kr_test = relative_permeabilities_twophase(relperm_params, Sw)
    for col in ("So", "Sw", "Sg", "kro", "krw", "krg"):
        assert col in df_kr_test.columns


def test_multiphase_flowproperties(relperm_params, df_pvt):
    p_frac = 1000
    p_res = pr
    phi = 0.1
    # scale pseudopressure
    df_pvt = rescale_pseudopressure(df_pvt, p_frac, p_res)
    df_kr = relative_permeabilities_twophase(relperm_params)
    reference_densities = {
        "rho_o0": 141.5 / (45 + 131.5),
        "rho_g0": 1.03e-3,
        "rho_w0": 1,
    }
    flow_props = FlowPropertiesTwoPhase.from_table(
        df_pvt, df_kr, reference_densities, phi, Sw, p_res
    )
    assert flow_props is not None, "Smoke testing FlowPropertiesTwoPhase"
    for fluid in "ogw":
        k_rel = flow_props.kr[f"kr{fluid}"]([0, 0.2, 0.4])
        assert np.max(k_rel) <= 1.0, "k_rel should be less than or equal to 1"
    # TODO: could use some more tests



"""Calculate fluid PVT properties."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass

import numpy as np
import pandas as pd
import scipy as sp
from numpy.typing import NDArray
from scipy.integrate import cumtrapz

from bluebonnet.fluids.gas import (
    b_factor_DAK,
    compressibility_DAK,
    density_DAK,
    make_nonhydrocarbon_properties,
    pseudocritical_point_Sutton,
    viscosity_Sutton,
    z_factor_DAK,
)
from bluebonnet.fluids.oil import (
    b_o_Standing,
    pressure_bubblepoint_Standing,
    viscosity_beggs_robinson,
)
from bluebonnet.fluids.water import b_water_McCain, viscosity_water_McCain

PRESSURE_STANDARD = 14.7
"""STP pressure (14.7 psi)"""
TEMPERATURE_STANDARD = 60
"""STP Temperature (60 degrees F)"""


def build_pvt_gas(
    gas_values: Mapping, gas_dryness: str, maximum_pressure: float = 14_000
) -> pd.DataFrame:
    """Build a table of PVT properties for use in the flow module.

    Parameters
    -----------
    gas_values : dictionary
        keys include 'N2','H2S','CO2', 'Gas Specific Gravity',
        'Reservoir Temperature (deg F)'
    gas_dryness : str
        One of 'wet gas' or 'dry gas'
    maximum_pressure : float
        initial reservoir pressure to calculate up to.

    Returns
    -------
    pd.DataFrame
        pvt_table
    """
    non_hydrocarbon_properties = make_nonhydrocarbon_properties(
        gas_values["N2"], gas_values["H2S"], gas_values["CO2"]
    )
    temperature_pc, pressure_pc = pseudocritical_point_Sutton(
        gas_values["Gas Specific Gravity"],
        non_hydrocarbon_properties,
        gas_dryness,
    )
    pressure = np.arange(10.0, maximum_pressure, 10.0)
    temperature = gas_values["Reservoir Temperature (deg F)"]
    z_factor = np.array(
        [z_factor_DAK(temperature, p, temperature_pc, pressure_pc) for p in pressure]
    )
    density = np.array(
        [
            density_DAK(
                temperature,
                p,
                temperature_pc,
                pressure_pc,
                float(gas_values["Gas Specific Gravity"]),
            )
            for p in pressure
        ]
    )
    viscosity = np.array(
        [
            viscosity_Sutton(
                temperature,
                p,
                temperature_pc,
                pressure_pc,
                float(gas_values["Gas Specific Gravity"]),
            )
            for p in pressure
        ]
    )
    compressibility = np.array(
        [
            compressibility_DAK(temperature, p, temperature_pc, pressure_pc)
            for p in pressure
        ]
    )
    pvt_gas = pd.DataFrame(
        data={
            "temperature": np.full_like(pressure, temperature),
            "pressure": pressure,
            "Density": density,
            "z-factor": z_factor,
            "compressibility": compressibility,
            "viscosity": viscosity,
        }
    )
    pseudopressure = 2 * cumtrapz(
        pvt_gas["pressure"] / (pvt_gas["viscosity"] * pvt_gas["z-factor"]),
        pvt_gas["pressure"],
        initial=0.0,
    )
    pvt_gas["pseudopressure"] = pseudopressure
    return pvt_gas


@dataclass
class Fluid:
    r"""Fluid PVT properties.

    Parameters
    ----------
    temperature: float
        reservoir temperature (in F)
    api_gravity: float
        oil gravity in degrees API
    gas_specific_gravity: float
        Gas gravity relative to air
    solution_gor_initial: float
        initial reservoir gas-oil ratio in scf/bbl
    salinity: float
        water salinity in weight percent total dissolved solids
    water_saturation_initial: float
        initial water saturation (V/V)
    """

    temperature: float
    """Temperature in F"""
    api_gravity: float
    """Oil API gravity"""
    gas_specific_gravity: float
    """Gas specific gravity"""
    solution_gor_initial: float
    """Initial gas-oil ratio"""
    salinity: float = 0.0
    """Salinity (total dissolved solids)"""
    water_saturation_initial: float = 0.0
    """Initial water saturation"""

    def water_FVF(
        self, pressure: NDArray[np.float] | float
    ) -> NDArray[np.float] | float:
        """Water formation volume factor (B-factor) from McCain.

        Parameters
        ----------
        pressure: np.ndarray
            water pressure in psia
        Returns
        ----------
        b_w: ndarray
            b-factor in reservoir bbl / standard bbl
        """
        b_w = np.array([b_water_McCain(self.temperature, p) for p in pressure])
        return b_w

    def water_viscosity(self, pressure: NDArray | float):
        """Water viscosity from McCain (1991).

        Parameters
        ----------
        pressure: ndarray
            water pressure in psia

        Returns
        -------
        mu_w : float
            viscosity in centipoise
        """
        return viscosity_water_McCain(self.temperature, pressure, self.salinity)

    def gas_FVF(
        self,
        pressure: NDArray | float,
        temperature_pseudocritical: float,
        pressure_pseudocritical: float,
    ) -> NDArray | float:
        """Gas formation volume factor (Bg) from Dranchuk and Abou-Kassem (1975).

        Parameters
        ----------
        pressure : ndarray
            reservoir pressure in psia
        temperature_pseudocritical : float
            pseudocritical temperature in Fahrenheit.
        pressure_pseudocritical : float
            pseudocritical pressure in psia

        Returns
        -------
        b_g: ndarray
            b-factor (reservoir barrels / scf)
        """
        b_g = np.array(
            [
                b_factor_DAK(
                    self.temperature,
                    p,
                    temperature_pseudocritical,
                    pressure_pseudocritical,
                )
                for p in pressure
            ]
        )
        return b_g

    def gas_viscosity(
        self,
        pressure: NDArray | float,
        temperature_pseudocritical: float,
        pressure_pseudocritical: float,
    ) -> NDArray | float:
        """Calculate the viscosity for gas using Sutton's Fudamental PVT Calculations (2007).

        Parameters
        ----------
        pressure : float
            reservoir pressure in psia
        temperature_pseudocritical : float
            pseudocritical temperature in Fahrenheit.
        pressure_pseudocritical : float
            pseudocritical pressure in psia

        Returns
        -------
        mu_g : ndarray
            viscosity_gas (centipoise)

        Examples
        -------
        >>> Fluid(400, 35, 0.65, 0).viscosity_gas(100, -102, 649)
        0.01652719692109309
        """
        viscosity = np.array(
            [
                viscosity_Sutton(
                    self.temperature,
                    p,
                    temperature_pseudocritical,
                    pressure_pseudocritical,
                    self.gas_specific_gravity,
                )
                for p in pressure
            ]
        )
        return viscosity

    def oil_FVF(self, pressure: NDArray | float) -> NDArray | float:
        """Calculate the oil formation volume factor (Bo) using Standing.

        Parameters
        ----------
        pressure : float
            reservoir pressure (psia)

        Returns
        -------
        b_o : ndarray
            :math:`b_o`, the formation volume factor (V/V)
        """
        fvf_oil = b_o_Standing(
            self.temperature,
            pressure,
            self.api_gravity,
            self.gas_specific_gravity,
            self.solution_gor_initial,
        )
        return fvf_oil

    def oil_viscosity(self, pressure: NDArray | float) -> NDArray | float:
        r"""Calculate the oil viscosity using Beggs-Robinson.

        Returns
        -------
        mu_o: ndarray
            :math:`\mu_o`, the oil viscosity
        """
        visc_br = np.vectorize(viscosity_beggs_robinson)
        mu_o = visc_br(
            self.temperature,
            pressure,
            self.api_gravity,
            self.gas_specific_gravity,
            self.solution_gor_initial,
        )
        return mu_o

    def pressure_bubblepoint(
        self,
    ) -> float:
        """Calculate the bubble point pressure using Standing.

        Returns
        -------
        p_bubble : float
            Bubble point pressure in psia.

        Examples
        -------
        >>> Fluid(200, 35, 0.8, 650).pressure_bubblepoint()
        2627.2017021875276
        """
        p_bubble = pressure_bubblepoint_Standing(
            self.temperature,
            self.api_gravity,
            self.gas_specific_gravity,
            self.solution_gor_initial,
        )
        return p_bubble


def pseudopressure(
    pressure: NDArray[np.float64],
    viscosity: NDArray[np.float64],
    z_factor: NDArray[np.float64],
) -> NDArray[np.float64]:
    """Calculate the pseudopressure using Al-Hussainy's relation.

    Parameters
    ----------
    pressure: NDArray, psi
    viscosity: NDArray, cp
    z_factor: NDArray, dimensionless

    Returns
    -------
    m : NDArray
        pseudopressure, psia^2 / cp
    """
    pp = 2 * pressure / (viscosity * z_factor)
    return sp.integrate.cumulative_trapezoid(pp, pressure, initial=0.0)



"""Flow properties for reservoir fluids.

Pressure (or pseudo-pressure) changes affect all sorts of things. This
module includes structures for storing these pressure-dependent properties
in order to aid the reservoir simulators in the `reservoir` module.
"""
from __future__ import annotations

import copy
import warnings
from collections import namedtuple
from collections.abc import Mapping

import numpy as np
import pandas as pd
from numpy import ndarray
from scipy.integrate import cumulative_trapezoid
from scipy.interpolate import LinearNDInterpolator, interp1d


class FlowProperties:
    r"""
    Flow properties for the system.

    This is used to translate from scaled pseudopressure to diffusivity and to capture
    the effect of expansion

    Attributes
    ----------
    pvt_props : Mapping
        has accessors for pseudopressure, alpha, compressibility, viscosity, z-factor

        * `pseudopressure`: pseudopressure in psi^2/centipoise
        * `pressure`: pore pressure
        * `alpha`: hydraulic diffusivity
        * `compressibility`: total compressibility
        * `viscosity`: dynamic viscosity :math:`\mu`
        * `z-factor`: compressibility factor, :math:`Z = p/\rho R T`
        * `m-scaled`: pseudopressure scaled by initial pseudopressure

    m_i : float
        pseudopressure that corresponds to initial reservoir pressure
    m_scaled_func: function
        calculates scaled pseudopressure from pressure
    alpha_func : function
        calculated hydraulic diffusivity from scaled pseudopressure
    """

    def __init__(self, pvt_props: Mapping[str, ndarray], p_i: float):
        """Wrap table of flow properties with useful methods.

        If alpha is not in the table, calculates hydraulic diffusivity as a function of
        compressibility and viscosity.

        Parameters
        -----------
        pvt_props : Mapping
            has accessors for pseudopressure, alpha (optional), compressibility, viscosity,
            z-factor

        p_i : float
            Initial reservoir pressure
        """
        pvt_props = copy.copy(pvt_props)
        need_cols_long = {
            "pseudopressure",
            "compressibility",
            "pressure",
            "viscosity",
            "z-factor",
        }
        need_cols_short = {"pressure", "pseudopressure", "alpha"}
        if (
            need_cols_long.intersection(pvt_props) != need_cols_long
            and need_cols_short.intersection(pvt_props) != need_cols_short
        ):
            raise ValueError("Need pvt_props to have: " + ", ".join(need_cols_short))
        if "alpha" in pvt_props:
            m_scale_func = interp1d(
                pvt_props["pressure"], 1 / pvt_props["pseudopressure"]
            )
            warnings.warn(
                "warning: scaling pseudopressure, using user's hydraulic diffusivity"
            )
            m_scaling_factor = m_scale_func(p_i)
        else:
            pseudopressure_scaling = (
                1
                / 2
                * pvt_props["compressibility"]
                * pvt_props["pressure"]
                * pvt_props["viscosity"]
                * pvt_props["z-factor"]
                / pvt_props["pressure"] ** 2
            )
            m_scale_func = interp1d(pvt_props["pressure"], pseudopressure_scaling)
            m_scaling_factor = m_scale_func(p_i)
            pvt_props["alpha"] = 1 / (  # mypy: ignore
                pvt_props["compressibility"] * pvt_props["viscosity"]
            )
        pvt_props["m-scaled"] = pvt_props["pseudopressure"] * m_scaling_factor
        self.m_scaled_func = interp1d(pvt_props["pressure"], pvt_props["m-scaled"])
        self.m_i = self.m_scaled_func(p_i)
        self.alpha = interp1d(
            pvt_props["m-scaled"],
            pvt_props["alpha"],
            fill_value=(min(pvt_props["alpha"]), max(pvt_props["alpha"])),
            bounds_error=False,
        )
        self.pvt_props = pvt_props

    def __repr__(self):
        """Representation."""
        return self.pvt_props.__repr__()


FlowPropertiesOnePhase = FlowProperties
"""Alias for :code:`FlowProperties`"""


class FlowPropertiesSimple(FlowProperties):
    """Flow properties where only viscosity and compressibility vary with pressure."""

    def __init__(self, pvt_props: Mapping[str, ndarray], p_i: float):
        """Wrap table of flow properties with useful methods.

        If alpha is not in the table, calculates hydraulic diffusivity as a function of
        compressibility and viscosity.

        Parameters
        -----------
        pvt_props : Mapping
            has accessors for pressure, compressibility, viscosity

        p_i : float
            Initial reservoir pressure
        """
        pvt_props = copy.copy(pvt_props)
        need_cols = {
            "compressibility",
            "pressure",
            "viscosity",
        }
        if need_cols.intersection(pvt_props) != need_cols:
            msg = "Need pvt_props to have: " + ", ".join(need_cols)
            raise ValueError(msg)
        pvt_props["alpha"] = 1 / (  # mypy: ignore
            pvt_props["compressibility"] * pvt_props["viscosity"]
        )
        pvt_props["m-scaled"] = pvt_props["pressure"]
        self.m_scaled_func = interp1d(pvt_props["pressure"], pvt_props["m-scaled"])
        self.m_i = self.m_scaled_func(p_i)
        self.alpha = interp1d(
            pvt_props["m-scaled"],
            pvt_props["alpha"],
            fill_value=(min(pvt_props["alpha"]), max(pvt_props["alpha"])),
            bounds_error=False,
        )
        self.pvt_props = pvt_props


class FlowPropertiesTwoPhase(FlowProperties):
    """
    Flow properties for the system.

    This is used to translate from scaled pseudopressure to diffusivity and to capture
    the effect of expansion

    References
    ----------
    Ruiz Maraggi, L.M., Lake, L.W. and Walsh, M.P., 2020. "A Two-Phase Non-Linear One-
    Dimensional Flow Model for Reserves Estimation in Tight Oil and Gas
    Condensate Reservoirs Using Scaling Principles." In SPE Latin American and
    Caribbean Petroleum Engineering Conference. OnePetro.
    https://doi.org/10.2118/199032-MS
    """

    @classmethod
    def from_table(
        cls,
        pvt_props: Mapping,
        kr_props: Mapping,
        reference_densities: dict,
        phi: float,
        Sw: float,
        p_i: float,
    ):
        """Create FlowProperties from tables.

        Using fluid properties from a PVT table and a relative permeability table,
        builds an interpolator for alpha from pseudopressure

        Parameters
        ----------
        df_pvt : Mapping
            table of PVT properties by pressure with co-varying So

            columns: pseudopressure, pressure, Bo, Bg, Bw, Rs, Rv, mu_o, mu_g, mu_w, So
        df_kr : Mapping
            table of relative permeabilities

            columns: So, Sg, Sw, kro, krg, krw
        reference_densities: dictionary of strings pointing to floats
            density at STP for oil, gas, and water
            rho_o0: oil, rho_g0: gas, rho_w0: water
        phi : float
            formation porosity (varies from 0 to 1)
        Sw : float
            water saturation (assumes Sw below or equal to Swirr)
        fvf_scale : float
            formation volume factor scale for recovery factor
        """
        need_cols_pvt = {
            "pseudopressure",
            "pressure",
            "Bo",
            "Bg",
            "Bw",
            "Rs",
            "Rv",
            "mu_o",
            "mu_g",
            "mu_w",
            "So",
        }
        need_cols_kr = {"So", "Sg", "Sw", "kro", "krg", "krw"}
        if need_cols_pvt.intersection(pvt_props) != need_cols_pvt:
            msg = f"df_pvt needs all of {need_cols_pvt}"
            raise ValueError(msg)
        if need_cols_kr.intersection(kr_props) != need_cols_kr:
            msg = f"df_kr needs all of {need_cols_kr}"
            raise ValueError(msg)
        pvt = {
            prop: interp1d(
                pvt_props["pressure"], pvt_props[prop], fill_value="extrapolate"
            )
            for prop in need_cols_pvt
        }
        pvt.update(reference_densities)
        # pvt.update({"rho_o0": rho_o0, "rho_g0": rho_g0, "rho_w0": rho_w0})
        kr = {
            fluid: interp1d(kr_props["So"], kr_props[fluid])
            for fluid in ("kro", "krg", "krw")
        }
        alpha_calc = alpha_multiphase(
            pvt_props["pressure"], pvt_props["So"], phi, Sw, pvt, kr
        )
        pseudopressure = pseudopressure_threephase(
            pvt_props["pressure"], pvt_props["So"], pvt, kr
        )
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            object = cls(
                {
                    "pressure": pvt_props["pressure"],
                    "pseudopressure": pseudopressure,
                    "alpha": alpha_calc,
                },
                p_i,
            )
        object.pvt = pvt
        object.kr = kr
        return object


def rescale_pseudopressure(
    df_pvt: Mapping[str, ndarray], p_frac: float, p_i: float
) -> Mapping[str, ndarray]:
    """Rescale pseudopressure to be 1 at p_i and 0 at p_frac.

    Parameters
    ----------
    df_pvt : Mapping[str,ndarray]
        table of PVT properties, including at least "pressure" and "pseudopressure"
    p_frac : float
        pressure at the frac face
    p_i : float
        initial reservoir pressure

    Returns
    -------
    df_pvt : Mapping[str,ndarray]
        new table of PVT properties
    """
    df_pvt = df_pvt.copy() if hasattr(df_pvt, "copy") else copy.deepcopy(df_pvt)
    pseudopressure = interp1d(df_pvt.pressure, df_pvt.pseudopressure)
    df_pvt["pseudopressure"] = (
        pseudopressure(df_pvt["pressure"]) - pseudopressure(p_frac)
    ) / (pseudopressure(p_i) - pseudopressure(p_frac))
    return df_pvt


def alpha_multiphase(
    pressure: ndarray, So: ndarray, phi: float, Sw: float, pvt: dict, kr: dict
) -> ndarray:
    """Calculate hydraulic diffusivity for a multiphase (two or three phase) system.

    Args
    ----
    pressure : ndarray
        absolute pressure for the cells
    So : ndarray
        Oil saturation for the cells
    phi : float
        porosity (assumed constant), in range from 0 to 1
    Sw : float
        Water saturation (assumed constant)
    pvt : dict
        PVT properties including

        * `rho_x0`: density at reference pressure
        * `mu_x`: viscosity (function of pressure)
        * `Bx`: formation volume factor (function of pressure) \
        for the x-components o (oil), g (gas), w (water)
        * `Rv`, `Rs`: dissolved and free gas (functions of pressure)
    kr : dict
        relative permeability for x-phases (o, g, w)

        `krx`: relative permeability (function of So)

    Returns
    -------
    alpha_t: ndarray
        total hydraulic diffusivity
    """
    lambda_combined = lambda_combined_func(pressure, So, pvt, kr)
    compressibility_combined = compressibility_combined_func(pressure, So, phi, Sw, pvt)
    return lambda_combined / compressibility_combined


def lambda_combined_func(
    pressure: ndarray, So: ndarray, pvt: dict, kr: dict
) -> ndarray:
    """Calculate mobility for three phase system.

    Args
    ----
    pressure : ndarray
        absolute pressure for the cells
    So : ndarray
        Oil saturation for the cells
    pvt : dict
        PVT properties including

        * `rho_x0`: density at reference pressure
        * `mu_x`: viscosity (function of pressure)
        * `Bx`: formation volume factor (function of pressure) \
        for the x-components o (oil), g (gas), w (water)
        * `Rv`, `Rs`: dissolved and free gas (functions of pressure)

    kr : dict
        relative permeability for x-phases (o, g, w)

        `krx`: relative permeability (function of So)

    Returns
    -------
    lambda_t: ndarray
        mobility for the cells
    """
    lambda_oil = pvt["rho_o0"] * (
        pvt["Rv"](pressure)
        * kr["krg"](So)
        / (pvt["mu_g"](pressure) * pvt["Bg"](pressure))
        + kr["kro"](So) / (pvt["mu_o"](pressure) * pvt["Bo"](pressure))
    )
    lambda_gas = pvt["rho_g0"] * (
        pvt["Rs"](pressure)
        * kr["kro"](So)
        / (pvt["mu_o"](pressure) * pvt["Bo"](pressure))
        + kr["krg"](So) / (pvt["mu_g"](pressure) * pvt["Bg"](pressure))
    )
    lambda_water = pvt["rho_w0"] * (
        kr["krw"](So) / (pvt["mu_w"](pressure) * pvt["Bw"](pressure))
    )  # + pvt["Rsw"](pressure) * kr["k_g"](So) / (pvt["mu_g"](pressure) * pvt["Bg"](pressure))
    return lambda_oil + lambda_gas + lambda_water


def compressibility_combined_func(
    pressure: ndarray, So: ndarray, phi: float, Sw: float | ndarray, pvt: dict
) -> ndarray:
    """Calculate three-phase compressibility.

    Args
    ----
    pressure : ndarray
        absolute pressure for the cells
    So : ndarray
        Oil saturation for the cells
    pvt : dict
        PVT properties including

        * `rho_x0`: density at reference pressure
        * `mu_x`: viscosity (function of pressure)
        * `Bx`: formation volume factor (function of pressure) \
        for the x-components o (oil), g (gas), w (water)
        * `Rv`, `Rs`: dissolved and free gas (functions of pressure)

    kr : dict
        relative permeability for x-phases (o, g, w)

        `krx`: relative permeability (function of So)

    Returns
    -------
    cp : ndarray
        Total fluid compressibility for the cells
    """
    Sg = 1 - So - Sw
    oil_cp = (
        phi
        * pvt["rho_o0"]
        * (
            pvt["Rv"](pressure + 0.5) * Sg / pvt["Bg"](pressure + 0.5)
            + So / pvt["Bo"](pressure + 0.5)
            - pvt["Rv"](pressure - 0.5) * Sg / pvt["Bg"](pressure - 0.5)
            + So / pvt["Bo"](pressure - 0.5)
        )
    )
    gas_cp = (
        phi
        * pvt["rho_g0"]
        * (
            pvt["Rs"](pressure + 0.5) * So / pvt["Bo"](pressure + 0.5)
            + Sg / pvt["Bg"](pressure + 0.5)
            - pvt["Rs"](pressure - 0.5) * So / pvt["Bo"](pressure - 0.5)
            + Sg / pvt["Bg"](pressure - 0.5)
        )
    )
    water_cp = (
        phi
        * pvt["rho_w0"]
        * (Sw / pvt["Bw"](pressure + 0.5) - Sw / pvt["Bw"](pressure - 0.5))
    )
    return oil_cp + gas_cp + water_cp


def pseudopressure_threephase(
    pressure: ndarray, So: ndarray, pvt: dict, kr: dict
) -> ndarray:
    """Calculate pseudopressure over pressure for three-phase fluid.

    Args
    ----
    pressure : ndarray
        absolute pressure for the cells
    So : ndarray
        Oil saturation for the cells
    pvt : dict
        PVT properties including

        * `rho_x0`: density at reference pressure
        * `mu_x`: viscosity (function of pressure)
        * `Bx`: formation volume factor (function of pressure) \
        for the x-components o (oil), g (gas), w (water)
        * `Rv`, `Rs`: dissolved and free gas (functions of pressure)

    kr : dict
        relative permeability for x-phases (o, g, w)

        `krx`: relative permeability (function of So)

    Returns
    -------
        pseudopressure : ndarray
    """
    lambda_oil = pvt["rho_o0"] * (
        pvt["Rv"](pressure)
        * kr["krg"](So)
        / (pvt["mu_g"](pressure) * pvt["Bg"](pressure))
        + kr["kro"](So) / (pvt["mu_o"](pressure) * pvt["Bo"](pressure))
    )
    lambda_gas = pvt["rho_g0"] * (
        pvt["Rs"](pressure)
        * kr["kro"](So)
        / (pvt["mu_o"](pressure) * pvt["Bo"](pressure))
        + kr["krg"](So) / (pvt["mu_g"](pressure) * pvt["Bg"](pressure))
    )
    lambda_water = pvt["rho_w0"] * (
        kr["krw"](So) / (pvt["mu_w"](pressure) * pvt["Bw"](pressure))
    )
    integrand = lambda_oil + lambda_gas + lambda_water
    pseudopressure = cumulative_trapezoid(pressure, integrand, initial=0)
    return pseudopressure


class FlowPropertiesMultiPhase(FlowProperties):
    """
    Flow properties for a multiphase system.

    This is used to translate from scaled pseudopressure and saturations to diffusivity
    and to capture the effect of expansion

    Parameters
    ----------
    df : Mapping
        has columns for pseudopressure, alpha, So, Sg, Sw

        * `pseudopressure`: pseudopressure scaled from 0 for frac face, 1 for initial\
            reservoir conditions
        * `alpha`: hydraulic diffusivity (needn't be scaled)
        * `So`: oil saturation
        * `Sg`: gas saturation
        * `Sw`: water saturation
    fvf_scale: dict
        includes Bo,Bg,Bw at initial conditions divided by FVF at the frac face
    """

    def __init__(self, df: Mapping[str, ndarray]):
        """Wrap table of pressure-dependent flow properties.

        Args:
            df (Mapping[str, ndarray]): table, including pseudopressure, So, Sg, Sw

        Raises:
            ValueError: Table columns are missing
        """
        need_cols = {"pseudopressure", "alpha", "So", "Sg", "Sw"}
        if need_cols.intersection(df.columns) != need_cols:
            msg = (
                "Need input dataframe to have 'pseudopressure', 'compressibility',"
                " and 'alpha' columns"
            )
            raise ValueError(msg)
        self.df = df
        x = df["pseudopressure", "So", "Sg", "Sw"]
        self.alpha = LinearNDInterpolator(x, df["alpha"])


RelPermParams = namedtuple(
    "RelPermParams", "n_o n_w n_g S_or S_wc S_gc k_ro_max k_rw_max k_rg_max"
)
"""Stores the parameters for a Brooks-Corey relative permeability model.

Parameters
----------
n_o : float
    Exponent for oil relative permeability
n_w : float
    Exponent for water relative permeability
n_g : float
    Exponent for gas relative permeability
S_or : float
    Residual saturation for oil (zero permeability point)
S_wc : float
    Residual saturation for water (zero permeability point)
S_gc : float
    Residual saturation for gas (zero permeability point)
k_ro_max : float
    Maximum relative permeability of the oil phase
k_rw_max : float
    Maximum relative permeability of the water phase
k_rg_max : float
    Maximum relative permeability of the gas phase
"""


def relative_permeabilities(
    saturations: ndarray,
    params: RelPermParams,
) -> ndarray:
    r"""Brooks-Corey power-law relative permeability.

    .. math ::
        k_{ro} = k_{ro,max}\left(\frac{S_o - S_{or}}{1-S_{or}-S_{wr}-S_{gr}}\right) \\
        k_{rw} = k_{rw,max}\left(\frac{S_w - S_{wr}}{1-S_{or}-S_{wr}-S_{gr}}\right) \\
        k_{rg} = k_{rg,max}\left(\frac{S_g - S_{gr}}{1-S_{or}-S_{wr}-S_{gr}}\right)

    Parameters
    ----------
    saturations: ndarray, specifically a record array
        records include So, Sg, Sw
    params: RelPermParams
        Includes Corey exponents, residual saturations, and max relative permeabilities

    Returns
    -------
    k_rel: numpy record array
        records include k_o, k_w, k_g (aka oil, water, gas)

    References
    ----------
    Brooks, R.H. and Corey, A.T. 1964. Hydraulic Properties of Porous Media.
    Hydrology Papers, No. 3, Colorado State U., Fort Collins, Colorado.

    https://petrowiki.spe.org/Relative_permeability_models
    """
    if np.any(np.abs([sum(v) - 1 for v in saturations]) > 1e-3):
        msg = "Saturations must sum to 1"
        raise ValueError(msg)
    if max(params.n_o, params.n_g, params.n_w) > 6:
        msg = "Exponents must be less than 6"
        raise ValueError(msg)
    if min(params.n_o, params.n_g, params.n_w) < 1:
        msg = "Exponents must be at least 1"
        raise ValueError(msg)
    if min(params.S_or, params.S_wc, params.S_gc) < 0:
        msg = "Critical saturations must be at least 0"
        raise ValueError(msg)
    if max(params.S_or, params.S_wc, params.S_gc) > 1:
        msg = "Critical saturations must be less than 1"
        raise ValueError(msg)
    if min(params.k_ro_max, params.k_rw_max, params.k_rg_max) < 0:
        msg = "Max relative permeability must be at least 0"
        raise ValueError(msg)
    if max(params.k_ro_max, params.k_rw_max, params.k_rg_max) > 1:
        msg = "Max relative permeability must be less than 1"
        raise ValueError(msg)

    denominator = 1 - params.S_or - params.S_wc - params.S_gc
    kro = (
        params.k_ro_max
        * ((saturations["So"] - params.S_or) / denominator) ** params.n_o
    )
    krw = (
        params.k_rw_max
        * ((saturations["Sw"] - params.S_wc) / denominator) ** params.n_w
    )
    krg = (
        params.k_rg_max
        * ((saturations["Sg"] - params.S_gc) / denominator) ** params.n_g
    )
    k_rel = np.array(
        list(zip(kro, krw, krg)),
        dtype=[(i, np.float64) for i in ("kro", "krw", "krg")],
    )
    for i in ("kro", "krw", "krg"):
        k_rel[i][k_rel[i] < 0] = 0  # negative permeability seems bad
    return k_rel


def relative_permeabilities_twophase(
    params: RelPermParams, Sw: float = 0.1
) -> pd.DataFrame:
    """Make two-phase relative permeability curves from Brooks-Corey.

    Parameters
    ----------
    params : RelPermParams
        Includes Corey exponents, residual saturations, and max relative permeabilities

    Returns
    -------
    df_kr: pd.DataFrame
        k_rx for saturations. Includes the columns "So","Sw","Sg","kro","krw","krg"

    Examples
    --------
    >>> relperm_params = RelPermParams(
            n_o=1, n_g=1, n_w=1, S_or=0, S_gc=0, S_wc=0.1, k_ro_max=1, k_rw_max=1, k_rg_max=1
        )
    >>> relative_permeabilities_twophase(relperm_params, 0.1)
    """
    if Sw > params.S_wc:
        msg = "Water saturation is above residual, so we have three flowing phases"
        raise ValueError(msg)
    saturations_test = pd.DataFrame(
        {
            "So": np.linspace(0, 1 - Sw, 50),
            "Sw": np.full(50, Sw),
            "Sg": np.linspace(1 - Sw, 0, 50),
        }
    )
    kr_matrix = pd.DataFrame(
        relative_permeabilities(saturations_test.to_records(index=False), params)
    )
    df_kr = pd.concat([saturations_test, kr_matrix], axis=1)
    return df_kr



"""Ease plotting production and fluid flow information."""
from __future__ import annotations

from typing import Union

import matplotlib.pyplot as plt
import matplotlib.scale as mscale
import matplotlib.ticker as ticker
import matplotlib.transforms as mtransforms
import numpy as np

from bluebonnet.flow import (
    IdealReservoir,
    MultiPhaseReservoir,
    SinglePhaseReservoir,
    TwoPhaseReservoir,
)

Reservoir = Union[
    IdealReservoir, SinglePhaseReservoir, TwoPhaseReservoir, MultiPhaseReservoir
]


class SquareRootScale(mscale.ScaleBase):
    """ScaleBase class for generating square root scale."""

    name = "squareroot"

    def __init__(self, axis, **kwargs):
        """Initialize for axis."""
        mscale.ScaleBase.__init__(self, axis, **kwargs)

    def set_default_locators_and_formatters(self, axis):
        """Set major and minor locators and formatters."""
        axis.set_major_locator(ticker.AutoLocator())
        axis.set_major_formatter(ticker.ScalarFormatter())
        axis.set_minor_locator(ticker.NullLocator())
        axis.set_minor_formatter(ticker.NullFormatter())

    def limit_range_for_scale(self, vmin, vmax, minpos):  # noqa: ARG
        """Do not allow negative values."""
        return max(0.0, vmin), vmax

    class SquareRootTransform(mtransforms.Transform):
        """Transform from linear to square root position."""

        input_dims = 1
        output_dims = 1
        is_separable = True

        def transform_non_affine(self, a):
            """Actual transform."""
            return np.array(a) ** 0.5

        def inverted(self):
            """Inverse transform."""
            return SquareRootScale.InvertedSquareRootTransform()

    class InvertedSquareRootTransform(mtransforms.Transform):
        """Inverted square-root transform."""

        input_dims = 1
        output_dims = 1
        is_separable = True

        def transform(self, a):
            """Square everything."""
            return np.array(a) ** 2

        def inverted(self):
            """Square root it. (Inverse of inverse)."""
            return SquareRootScale.SquareRootTransform()

    def get_transform(self):
        """Get square root transform."""
        return self.SquareRootTransform()


mscale.register_scale(SquareRootScale)


def plot_pseudopressure(
    reservoir: Reservoir,
    every: int = 200,
    rescale: bool = False,
    ax: plt.Axes = None,
    x_max: float = 1,
    y_max: float | None = None,
) -> plt.Axes:
    """Plot pseudopressure versus distance over time.

    Parameters
    ----------
    reservoir : Reservoir
        simulated reservoir
    every : int, optional
        timesteps between a plotted pseudopressure profile, by default 200
    rescale : bool, optional
        if true, rescale pseudopressure by initial value, by default False
    ax : plt.Axes, optional
        axes to plot on, by default None
    x_max : float, optional
        maximum distance to plot, by default 1
    y_max : float | None, optional
        maximum pseudopressure to plot, by default None

    Returns
    -------
    plt.Axes
        pseudopressure plotted
    """
    if ax is None:
        _, ax = plt.subplots()
    x = np.linspace(1 / reservoir.nx, 1, reservoir.nx)
    pinit = reservoir.pseudopressure[0, -1]
    for i, p in enumerate(reservoir.pseudopressure):
        if i % every == 0:
            if rescale:
                pscale = (p - p[0]) / (pinit - p[0])
                ax.plot(x, pscale, color="steelblue")
            else:
                ax.plot(x, p, color="steelblue")
    ax.set(xlabel="x", ylabel="Pseudopressure", xlim=(0, x_max), ylim=(0, y_max))
    return ax


def plot_recovery_rate(
    reservoir: Reservoir, ax: plt.Axes | None = None, change_ticks: bool = False
) -> plt.Axes:
    """Plot recovery rate over time.

    Parameters
    ----------
    reservoir : Reservoir
        simulated reservoir
    ax : plt.Axes | None, optional
        axes to plot on, by default None
    change_ticks : bool, optional
        if true, set xticks, by default False

    Returns
    -------
    plt.Axes
        recovery rate, plotted
    """
    if ax is None:
        _, ax = plt.subplots()

    cumulative = reservoir.recovery_factor()
    rate = np.gradient(cumulative, reservoir.time)
    ax.plot(reservoir.time, rate)
    ax.set(
        xscale="log",
        yscale="log",
        ylim=(1.0e-4, None),
        xlim=(1e-6, max(reservoir.time)),
        xlabel="Scaled time",
        ylabel="Recovery rate",
    )
    if change_ticks:
        tick_locs = (np.logspace(-7, 0, 7) * np.sqrt(max(reservoir.time))) ** 2
        ax.set_xticks(tick_locs)
    return ax


def plot_recovery_factor(
    reservoir: Reservoir, ax: plt.Axes | None = None, change_ticks: bool = False
) -> plt.Axes:
    """Plot cumulatie recovery over time.

    Parameters
    ----------
    reservoir : Reservoir
        simulated reservoir
    ax : plt.Axes | None, optional
        axes to plot on, by default None
    change_ticks : bool, optional
        if true, set xticks, by default False

    Returns
    -------
    plt.Axes
        recovery rate, plotted
    """
    if ax is None:
        _, ax = plt.subplots()
    rf = reservoir.recovery_factor()
    time = reservoir.time
    ax.plot(time, rf)
    ax.set(
        xscale="squareroot",
        ylim=(0, None),
        xlim=(0, max(time)),
        xlabel="Scaled time",
        ylabel="Recovery factor",
    )
    if change_ticks:
        tick_locs = np.round((np.linspace(0, 1, 7) * np.sqrt(max(time))) ** 2, 1)
        ax.set_xticks(tick_locs)
    return ax



"""Define a suite a tests for the reservoir module."""
from __future__ import annotations

from itertools import product

import numpy as np
import pandas as pd
import pytest
from bluebonnet.flow import (
    FlowProperties,
    IdealReservoir,
    MultiPhaseReservoir,
    SinglePhaseReservoir,
)
from scipy.optimize import curve_fit

nx = (30,)
nt = (1200,)
pf = (100,)
pr = (8_000,)
columns_renamer_gas = {
    "P": "pressure",
    "Z-Factor": "z-factor",
    "Cg": "compressibility",
    "Viscosity": "viscosity",
    "Density": "density",
}
columns_renamer_oil = {
    "P": "pressure",
    "Z-Factor": "z-factor",
    "Co": "compressibility",
    "Oil_Viscosity": "viscosity",
    "Oil_Density": "density",
}
pvt_gas = pd.read_csv("tests/data/pvt_gas.csv").rename(columns=columns_renamer_gas)
pvt_oil = pd.read_csv("tests/data/pvt_oil.csv").rename(columns=columns_renamer_oil)
fluid = (FlowProperties(pvt_gas, pr[-1]),)  # FlowProperties(pvt_oil, pr[-1]))
sim_props = list(product(nx, pf, pr, fluid))
So, Sg, Sw = (0.7, 0.2, 0.1)
reservoirs = (IdealReservoir, SinglePhaseReservoir)  # TODO:, MultiPhaseReservoir)


@pytest.mark.parametrize("Reservoir", reservoirs)
@pytest.mark.parametrize(("nx", "pf", "pi", "fluid"), sim_props)
def reservoir_start(nx, pf, pi, fluid, Reservoir):
    if Reservoir == MultiPhaseReservoir:
        reservoir = Reservoir(nx, pf, pi, fluid, So, Sg, Sw)
        assert reservoir.So == So
        assert reservoir.Sg == Sg
        assert reservoir.Sw == Sw
    else:
        reservoir = Reservoir(nx, pf, pi, fluid)

    assert reservoir.nx == nx
    assert reservoir.pressure_fracface == pf
    assert reservoir.pressure_initial == pi
    assert reservoir.fluid == fluid


@pytest.mark.parametrize("nt", nt)
@pytest.mark.parametrize("Reservoir", reservoirs)
@pytest.mark.parametrize("nx,pf,pi,fluid", sim_props)
class TestRun:
    def test_rf_fails_early(self, nx, pf, pi, fluid, Reservoir, nt):  # noqa: ARG002
        if Reservoir == MultiPhaseReservoir:
            reservoir = Reservoir(nx, pf, pi, fluid, So, Sg, Sw)
        else:
            reservoir = Reservoir(nx, pf, pi, fluid)
        with pytest.raises(RuntimeError):
            # make sure you can't run recovery_factor before simulate
            reservoir.recovery_factor()

    def test_rf_early_asymptotes(self, nx, pf, pi, fluid, Reservoir, nt):
        if Reservoir == MultiPhaseReservoir:
            reservoir = Reservoir(nx, pf, pi, fluid, So, Sg, Sw)
        else:
            reservoir = Reservoir(nx, pf, pi, fluid)
        end_t = 9.0
        time = np.linspace(0, np.sqrt(end_t), nt) ** 2
        reservoir.simulate(time)

        def logslope(time, rf, mask=None):
            if mask is None:
                mask = (time > 0.02) & (time < 0.3)
            time_log = np.log(time[mask])
            rf_log = np.log(rf[mask])

            def curve(time, intercept, slope):
                return intercept + time * slope

            (intercept, slope), _ = curve_fit(curve, time_log, rf_log, p0=[0, 1])
            return slope

        # rf goes as sqrt-time in early production
        for density in (True, False):
            rf = reservoir.recovery_factor(density=density)
            slope = logslope(time, rf, (time > 0.1) & (time < 0.3))
            assert (
                np.abs(slope - 0.5) < 0.05
            ), "Early recovery factor goes as the square root"

    def test_rf_late_asymptotes(self, nx, pf, pi, fluid, Reservoir, nt):
        if Reservoir == MultiPhaseReservoir:
            reservoir = Reservoir(nx, pf, pi, fluid, So, Sg, Sw)
        else:
            reservoir = Reservoir(nx, pf, pi, fluid)
        end_t = 100
        time = np.linspace(0, np.sqrt(end_t), nt) ** 2
        reservoir.simulate(time)
        if False:
            import matplotlib.pyplot as plt
            from bluebonnet.plotting import plot_pseudopressure, plot_recovery_factor

            fig, ax = plt.subplots()
            ax = plot_recovery_factor(reservoir, ax)
            fig.savefig("rf.png")
            fig, ax = plt.subplots()
            ax = plot_pseudopressure(reservoir, 100, ax=ax)
            fig.savefig("pseudopressure.png")

        def logslope(time, rf, mask=None):
            if mask is None:
                mask = (time > 0.02) & (time < 0.2)
            time_log = np.log(time[mask])
            rf_log = np.log(rf[mask])

            def curve(time, intercept, slope):
                return intercept + time * slope

            (_, slope), _ = curve_fit(curve, time_log, rf_log, p0=[0, 1])
            return slope

        # rf stops increasing at late time
        for density in (True, False):
            rf = reservoir.recovery_factor(density=density)
            slope = logslope(time, rf, time > 10)
            assert np.abs(slope) < 0.03, "Late recovery factor is slow"
            assert slope > -1e-10, "Late recovery never trends negative"



"""Oil PVT properties."""

from __future__ import annotations

import math

import numpy as np
from numpy.typing import NDArray

from bluebonnet.fluids.gas import b_factor_DAK


def b_o_Standing(
    temperature: float,
    pressure: NDArray | float,
    api_gravity: float,
    gas_specific_gravity: float,
    solution_gor_initial: float,
) -> NDArray | float:
    """Calculate the oil formation volume factor (Bo) using Standing.

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    pressure : float
        reservoir pressure (psia)
    api_gravity : float
        Oil gravity in degrees API.
    gas_specific_gravity : float
        Gas gravity relative to air.
    solution_gor_initial : float
        Initial gas:oil ratio in scf/bbl.

    Returns
    -------
    float
        Bo, the formation volume factor (V/V)

    Examples
    -------
    >>> b_o_Standing(200, 3_000, 35, 0.8, 650)

    >>> b_o_Standing(200, 2_000, 35, 0.8, 650)
    1.2820114682225974
    """
    pressure_bubblepoint = pressure_bubblepoint_Standing(
        temperature, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    fvf_bubblepoint = b_o_bubblepoint_Standing(
        temperature, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    solution_gor = solution_gor_Standing(
        temperature, pressure, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    if np.ndim(pressure) != 0:  # check for non-scalar
        compressibility_undersat = oil_compressibility_undersat_Spivey(
            temperature,
            pressure[pressure >= pressure_bubblepoint],
            api_gravity,
            gas_specific_gravity,
            solution_gor_initial,
        )
        fvf_oil = np.empty_like(pressure)
        fvf_oil[pressure >= pressure_bubblepoint] = fvf_bubblepoint * np.exp(
            compressibility_undersat
            * (pressure_bubblepoint - pressure[pressure >= pressure_bubblepoint])
        )
        fvf_oil[pressure < pressure_bubblepoint] = b_o_bubblepoint_Standing(
            temperature,
            api_gravity,
            gas_specific_gravity,
            solution_gor[pressure < pressure_bubblepoint],
        )
    else:
        if pressure >= pressure_bubblepoint:
            compressibility_undersat = oil_compressibility_undersat_Spivey(
                temperature,
                pressure,
                api_gravity,
                gas_specific_gravity,
                solution_gor_initial,
            )
            fvf_oil = fvf_bubblepoint * np.exp(
                compressibility_undersat * (pressure_bubblepoint - pressure)
            )
        else:
            fvf_oil = b_o_bubblepoint_Standing(
                temperature, api_gravity, gas_specific_gravity, solution_gor
            )
    return fvf_oil


def pressure_bubblepoint_Standing(
    temperature: float,
    api_gravity: float,
    gas_specific_gravity: float,
    solution_gor_initial: float,
) -> float:
    """Calculate the bubble point pressure using Standing.

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    api_gravity : float
        Oil gravity in degrees API.
    gas_specific_gravity : float
        Gas gravity relative to air.
    solution_gor_initial : float
        Gas: oil ratio in scf/bbl.

    Returns
    -------
    p_b : float
        Bubble point pressure in psia.

    Examples
    -------
    >>> pressure_bubblepoint_Standing(200, 35, 0.8, 650)
    2627.2017021875276
    """
    pressure_bubble = 18.2 * (
        (solution_gor_initial / gas_specific_gravity) ** 0.83
        * 10 ** (0.00091 * temperature - 0.0125 * api_gravity)
        - 1.4
    )
    return pressure_bubble


def b_o_bubblepoint_Standing(
    temperature: float,
    api_gravity: float,
    gas_specific_gravity: float,
    solution_gor_initial: NDArray | float,
) -> NDArray | float:
    """Calculate the oil formation volume factor (Bo) at the bubble point using Standing.

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    api_gravity : float
        Oil gravity in degrees API.
    gas_specific_gravity : float
        Gas gravity relative to air.
    solution_gor_initial : NDArray
        Initial gas:oil ratio in scf/bbl.

    Returns
    -------
    b_o : NDArray
        Oil formation volume factor (rb/stb)

    Example
    -------
    >>> b_o_bubblepoint_Standing(200, 35, 0.8, 650)
    1.3860514623492897
    """
    oil_specific_gravity = 141.5 / (131.5 + api_gravity)
    fvf_bubblepoint = (
        0.9759
        + 0.00012
        * (
            solution_gor_initial * np.sqrt(gas_specific_gravity / oil_specific_gravity)
            + 1.25 * temperature
        )
        ** 1.2
    )
    return fvf_bubblepoint


def db_o_dgor_Standing(
    temperature: float,
    api_gravity: float,
    gas_specific_gravity: float,
    solution_gor_initial: float,
) -> float:
    """Calculate the derivative of Bo per solution GOR at the bubble point.

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    api_gravity : float
        Oil gravity in degrees API.
    gas_specific_gravity : float
        Gas gravity relative to air.
    solution_gor_initial : NDArray
        Initial gas:oil ratio in scf/bbl.

    Returns
    -------
    dbo_dgor:  NDArray
        Change in Oil formation volume factor per unit GOR (rb/stb / (scf/bbl))

    Example
    -------
    >>> db_o_dgor_Standing(200, 35, 0.8, 650)
    """
    oil_specific_gravity = 141.5 / (131.5 + api_gravity)
    sqrt_gravity_ratio = np.sqrt(gas_specific_gravity / oil_specific_gravity)
    dB_o_dgor = (
        1.2
        * 1.2e-4
        * (solution_gor_initial * sqrt_gravity_ratio + 1.25 * temperature) ** 0.2
        * sqrt_gravity_ratio
    )
    return dB_o_dgor


def solution_gor_Standing(
    temperature: float,
    pressure: NDArray | float,
    api_gravity: float,
    gas_specific_gravity: float,
    solution_gor_initial: float,
) -> NDArray | float:
    """Calculate the solution GOR partition using Standing.

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    pressure : float or array
        reservoir pressure (psia)
    api_gravity : float
        Oil gravity in degrees API.
    gas_specific_gravity : float
        Gas gravity relative to air.
    solution_gor_initial : float
        Gas: oil ratio in scf/bbl.

    Returns
    -------
    gor : float
        Solution GOR in scf/stb

    Examples
    -------
    >>> solution_gor_Standing(200, 3_000, 35, 0.8, 650)
    650
    >>> solution_gor_Standing(200, 2_000, 35, 0.8, 650)
    453.6099792270382
    """
    pressure_bubblepoint = pressure_bubblepoint_Standing(
        temperature, api_gravity, gas_specific_gravity, solution_gor_initial
    )

    def gor_belowbubble(pressure: NDArray | float) -> NDArray | float:
        gor = gas_specific_gravity * (
            (pressure / 18.2 + 1.4)
            * 10 ** (0.0125 * api_gravity - 0.00091 * temperature)
        ) ** (1 / 0.83)
        return gor

    if np.ndim(pressure) != 0:
        solution_gor = np.full_like(pressure, solution_gor_initial)
        solution_gor[pressure < pressure_bubblepoint] = gor_belowbubble(
            pressure[pressure < pressure_bubblepoint]
        )
    else:
        if pressure >= pressure_bubblepoint:
            solution_gor = solution_gor_initial
        else:
            solution_gor = gor_belowbubble(pressure)
    return solution_gor


def dgor_dpressure_Standing(
    temperature: float,
    pressure: float,
    api_gravity: float,
    gas_specific_gravity: float,
    solution_gor_initial: float,
) -> float:
    r"""Calculate the instantaneous change in GOR with pressure using Standing.

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    pressure : float
        reservoir pressure (psia)
    api_gravity : float
        Oil gravity in degrees API.
    gas_specific_gravity : float
        Gas gravity relative to air.
    solution_gor_initial : float
        Initial gas:oil ratio in scf/bbl.

    Returns
    -------
    dgor : float
        Change in GOR per pressure ( scf / scf / psi). :math:`\partial R_s/\partial p`

    Examples
    --------
    >>> dgor_dpressure_Standing(200, 3_000, 35, 0.8, 650)
    0
    >>> dgor_dpressure_Standing(200, 2_000, 35, 0.8, 650)
    0.2824395998221715
    """
    pressure_bubblepoint = pressure_bubblepoint_Standing(
        temperature, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    if pressure >= pressure_bubblepoint:
        d_gor = 0.0
    else:
        d_gor = (
            gas_specific_gravity
            / (0.83 * 18.2)
            * (pressure / 18.2 + 1.4) ** (1 / 0.83 - 1)
            * (10 ** ((0.0125 * api_gravity - 0.00091 * temperature) / 0.83))
        )
    return d_gor


def oil_compressibility_undersat_Standing(
    temperature: float,
    pressure: float,
    api_gravity: float,
    gas_specific_gravity: float,
    solution_gor_initial: float,
) -> float:
    """Calculate the oil compressibility (c_o) using Standing.

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    pressure : float
        reservoir pressure (psia)
    api_gravity : float
        Oil gravity in degrees API.
    gas_specific_gravity : float
        Gas gravity relative to air.
    solution_gor_initial : float
        Initial gas:oil ratio in scf/bbl.

    Returns
    -------
    c_o : float
        oil compressibility in 1/psi

    Example
    -------
    >>> oil_compressibility_undersat_Standing(200, 3_000, 35, 0.8, 650)
    1.478788544207282e-05
    """
    oil_specific_gravity = 141.5 / (131.5 + api_gravity)
    pressure_bubblepoint = pressure_bubblepoint_Standing(
        temperature, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    b_o_bubblepoint = b_o_bubblepoint_Standing(
        temperature, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    density_bubblepoint = (
        62.37 * oil_specific_gravity
        + 0.0136 * gas_specific_gravity * solution_gor_initial
    ) / b_o_bubblepoint
    oil_compressibility = 1e-6 * math.exp(
        (density_bubblepoint + 0.004347 * (pressure - pressure_bubblepoint) - 79.1)
        / (7.141e-4 * (pressure - pressure_bubblepoint) - 12.938)
    )
    return oil_compressibility


def oil_compressibility_undersat_Spivey(
    temperature: float,
    pressure: NDArray | float,
    api_gravity: float,
    gas_specific_gravity: float,
    solution_gor_initial: float,
) -> NDArray | float:
    """Calculate the oil compressibility (c_o) using Spivey.

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    pressure : float or array
        reservoir pressure (psia)
    api_gravity : float
        Oil gravity in degrees API.
    gas_specific_gravity : float
        Gas gravity relative to air.
    solution_gor_initial : float
        Initial gas:oil ratio in scf/bbl.

    Returns
    -------
    c_o : float
        oil compressibility in 1/psi

    Examples
    --------
    >>> oil_compressibility_undersat_Spivey(200, 3_000, 35, 0.8, 650)
    9.177554347361075e-06
    """
    pressure_bubblepoint = pressure_bubblepoint_Standing(
        temperature, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    # sometimes this is passed an empty pressure array when pressures are all
    # above the bubblepoint
    if np.size(pressure) == 0:
        return np.array([], dtype=np.float64)
    reduced_pressure = pressure / pressure_bubblepoint
    # set up constants for quadratic formula
    # C_labels = [
    #     'api_gravity', 'gas_specific_gravity', 'pressure_bubblepoint',
    #     'reduced_pressure', 'solution_gor_initial','temperature'
    # ]
    C0 = np.array([3.011, -0.835, 3.51, 0.327, -1.918, 2.52])
    C1 = np.array([-2.6254, -0.259, -0.0289, -0.608, -0.642, -2.73])
    C2 = np.array([0.497, 0.382, -0.0584, 0.0911, 0.154, 0.429])
    if np.ndim(reduced_pressure) == 0:
        X = np.log(
            [
                api_gravity,
                gas_specific_gravity,
                pressure_bubblepoint,
                reduced_pressure,
                solution_gor_initial,
                temperature,
            ],
            dtype="f8",
        )
    else:
        X = np.log(
            [
                [
                    api_gravity,
                    gas_specific_gravity,
                    pressure_bubblepoint,
                    rp,
                    solution_gor_initial,
                    temperature,
                ]
                for rp in reduced_pressure
            ]
        )
    z = np.sum(C0) + X @ C1 + (X**2) @ C2
    compressibility_bubblepoint = np.exp(2.434 + 0.475 * z + 0.048 * z**2)
    compressibility_derivative = (-0.608 + 0.1822 * np.log(reduced_pressure)) / pressure
    compressibility_2nd_derivative = (
        compressibility_bubblepoint * (0.475 + 0.096 * z) * compressibility_derivative
    )
    oil_compressibility = 1e-6 * (
        compressibility_bubblepoint
        + (pressure - pressure_bubblepoint) * compressibility_2nd_derivative
    )
    return oil_compressibility


def oil_compressibility_Standing(
    temperature: float,
    pressure: float,
    api_gravity: float,
    gas_specific_gravity: float,
    solution_gor_initial: float,
    temperature_pseudocritical: float,
    pressure_pseudocritical: float,
    temperature_standard: float = 60,
    pressure_standard: float = 14.7,
) -> float:
    """Calculate the oil formation volume factor (Bo) using Standing.

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    pressure : float
        reservoir pressure (psia)
    api_gravity : float
        Oil gravity in degrees API.
    gas_specific_gravity : float
        Gas gravity relative to air.
    solution_gor_initial : float
        Initial gas:oil ratio in scf/bbl.
    pressure_pseudocritical : float
        Gas pseudocritical pressure (use gas module) in psia
    temperature_pseudocritical : float
        Gas pseudocritical temperature (use gas module) in F
    temperature_standard : float
        standard temperature in F (default: 60)
    pressure_standard : float
        standard pressure in psia (default: 14.70)

    Returns
    -------
    c_o : float
        oil compressibility in 1/psi

    Examples
    ------
    >>> oil_compressibility_Standing(200, 3000, 35, 0.8, 650, -72.2, 653)

    >>> oil_compressibility_Standing(200, 2000, 35, 0.8, 650, -72.2, 653)

    """
    pressure_bp = pressure_bubblepoint_Standing(
        temperature, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    if pressure >= pressure_bp:
        compressibility = oil_compressibility_undersat_Spivey(
            temperature,
            pressure,
            api_gravity,
            gas_specific_gravity,
            solution_gor_initial,
        )
    else:
        b_g = b_factor_DAK(
            temperature,
            pressure,
            temperature_pseudocritical,
            pressure_pseudocritical,
            temperature_standard,
            pressure_standard,
        )
        solution_gor = solution_gor_Standing(
            temperature,
            pressure,
            api_gravity,
            gas_specific_gravity,
            solution_gor_initial,
        )
        dGOR_dpressure = (
            gas_specific_gravity
            / 0.83
            * (pressure / 18.2 + 1.4) ** (1 / 0.83 - 1)
            / 18.2
            * 10 ** ((0.0125 * api_gravity - 0.00091 * temperature) / 0.83)
        )
        b_o_bubblepoint = b_o_bubblepoint_Standing(
            temperature, api_gravity, gas_specific_gravity, solution_gor_initial
        )
        dBo_dGOR = db_o_dgor_Standing(
            temperature, api_gravity, gas_specific_gravity, solution_gor
        )
        compressibility = (b_g - dBo_dGOR) * dGOR_dpressure / b_o_bubblepoint
    return compressibility


def density_Standing(
    temperature: float,
    pressure: float,
    api_gravity: float,
    gas_specific_gravity: float,
    solution_gor_initial: float,
) -> float:
    """Calculate the oil density using Standing.

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    pressure : float
        reservoir pressure (psia)
    api_gravity : float
        Oil gravity in degrees API.
    gas_specific_gravity : float
        Gas gravity relative to air.
    solution_gor_initial : float
        Initial gas:oil ratio in scf/bbl.

    Returns
    -------
    rho_o : float
        oil density (lb/ft^3)

    Examples
    ------
    >>> density_Standing(200, 2000, 35, 0.8, 650)
    44.989365953905136
    """
    oil_specific_gravity = 141.5 / (131.5 + api_gravity)
    solution_gor = solution_gor_Standing(
        temperature, pressure, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    b_o = b_o_Standing(
        temperature, pressure, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    density = (
        62.37 * oil_specific_gravity + 0.0136 * gas_specific_gravity * solution_gor
    ) / b_o
    return density


def viscosity_beggs_robinson(
    temperature: float,
    pressure: float,
    api_gravity: float,
    gas_specific_gravity: float,
    solution_gor_initial: float,
) -> float:
    r"""Calculate the oil viscosity using Beggs-Robinson.

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    pressure : float
        reservoir pressure (psia)
    api_gravity : float
        Oil gravity in degrees API.
    gas_specific_gravity : float
        Gas gravity relative to air.
    solution_gor_initial : float
        Initial gas:oil ratio in scf/bbl.

    Returns
    -------
    mu_o : float
        :math:`\mu_o`, the oil viscosity
    """
    pressure_bubblepoint = pressure_bubblepoint_Standing(
        temperature, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    solution_gor = solution_gor_Standing(
        temperature, pressure, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    if pressure >= pressure_bubblepoint:
        mu_o_dead = (
            10 ** (10 ** (3.0324 - 0.02023 * api_gravity) * temperature**-1.163) - 1
        )
        mu_o_live = _mu_dead_to_live_br(mu_o_dead, solution_gor_initial)
        mu_o = mu_o_live * (pressure / pressure_bubblepoint) ** (
            2.6 * pressure**1.187 * np.exp(-11.513 - 8.98e-5 * pressure)
        )
    else:
        mu_o_dead = (
            10 ** (10 ** (3.0324 - 0.02023 * api_gravity) * temperature**-1.163) - 1
        )
        mu_o = _mu_dead_to_live_br(mu_o_dead, solution_gor)
    return mu_o


def _mu_dead_to_live_br(
    mu_dead: NDArray | float, solution_gor_initial: NDArray | float
) -> NDArray | float:
    mu_live = (
        10.715
        * (solution_gor_initial + 100) ** -0.515
        * mu_dead ** (5.44 * (solution_gor_initial + 150) ** -0.338)
    )
    return mu_live



"""Gas pvt properties, using Dranchuk and Abou-Kassem's correlations."""
from __future__ import annotations

import math

import numpy as np
from numpy.typing import NDArray
from scipy.integrate import quad
from scipy.optimize import minimize


# Gas property calculations
def make_nonhydrocarbon_properties(
    nitrogen: float, hydrogen_sulfide: float, co2: float, *others: dict[str, float]
) -> NDArray:
    """Create an array of the nonhydrocarbon molecules present.

    Parameters
    ----------
    nitrogen : float
        compositional fraction of N2
    hydrogen_sulfide : float
        compositional fraction of H2S
    co2 : float
        compositional fraction of CO2
    *others : list(tuple)
       list of tuples of (name, compositional fraction, molecular weight,
                          critical temperature (in R), critical pressure (psia))
       for other non-hydrocarbon molecules

    Returns
    --------
    non_hydrocrabon_properties : NDArray
        structured array of non-hydrocarbon fluid properties

    Examples
    --------
    >>> make_nonhydrocarbon_properties(0.03, 0.012, 0.018)
    """
    non_hydrocarbon_properties = np.array(
        [
            ("Nitrogen", nitrogen, 28.01, 226.98, 492.26),
            ("Hydrogen sulfide", hydrogen_sulfide, 34.08, 672.35, 1299.97),
            ("CO2", co2, 44.01, 547.54, 1070.67),
            *others,
        ],
        dtype=[
            ("name", "U20"),
            ("fraction", "f8"),
            ("molecular weight", "f8"),
            ("critical temperature", "f8"),
            ("critical pressure", "f8"),
        ],
    )
    return non_hydrocarbon_properties


def z_factor_DAK(
    temperature: float,
    pressure: float,
    temperature_pseudocritical: float,
    pressure_pseudocritical: float,
) -> float:
    """Calculate the z-factor for gas using Dranchuk and Abou-Kassem (1975).

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    pressure : float
        reservoir pressure in psia
    temperature_pseudocritical : float
        pseudocritical temperature in Fahrenheit.
    pressure_pseudocritical : float
        pseudocritical pressure in psia

    Returns
    -------
    z_fact: float
        z_factor (dimensionless)

    Examples
    ------
    >>> z_factor_DAK(400, 100, -102, 649)
    0.9969013621293381
    """
    A = np.array(
        [
            0.3265,
            -1.07,
            -0.5339,
            0.01569,
            -0.05165,
            0.5475,
            -0.7361,
            0.1844,
            0.1056,
            0.6134,
            0.721,
        ]
    )
    temp_reduced = (temperature + 459.67) / (temperature_pseudocritical + 459.67)
    pressure_reduced = pressure / pressure_pseudocritical
    C = np.zeros(5)  # Taylor series expansion
    C[0] = (
        A[0] * A[1] / temp_reduced
        + A[2] / (temp_reduced**3)
        + A[3] / (temp_reduced**4)
        + A[4] / (temp_reduced**5)
    )
    C[1] = A[5] + A[6] / temp_reduced + A[7] / (temp_reduced**2)
    C[2] = -A[8] * (A[6] / temp_reduced + A[7] / (temp_reduced**2))
    C[3] = A[9] / (temp_reduced**3)
    C[4] = A[9] * A[10] / (temp_reduced**3)

    def calculate_error_fraction(rho: NDArray[np.float64]):
        rho = rho[0]
        B = math.exp(-A[10] * rho**2)
        F_rho = (
            0.27 * pressure_reduced / (temp_reduced * rho)
            - 1
            - C[0] * rho
            - C[1] * rho**2
            - C[2] * rho**5
            - C[3] * rho**2 * B
            - C[4] * rho**4 * B
        )
        DF_rho = (
            -0.27 * pressure_reduced / (temp_reduced * rho**2)
            - C[0]
            - 2 * C[1] * rho
            - 5 * C[2] * rho**4
            - 2 * C[3] * rho * B
            + 2 * A[10] * rho * B * C[3] * rho**2
            - 4 * C[4] * rho**3 * B
            + 2 * A[10] * rho * B * C[4] * rho**4
        )
        return math.fabs(F_rho / DF_rho)

    rho_guess = 0.27 * pressure_reduced / temp_reduced
    bounds = (
        (rho_guess / 5, rho_guess * 20),
    )  # bounds go from a z-factor of 0.05 to 5
    result = minimize(calculate_error_fraction, rho_guess, bounds=bounds)
    rho = result.x[0]
    Z_factor = 0.27 * pressure_reduced / (rho * temp_reduced)
    return Z_factor


def z_factor_hallyarbrough(pressure: float, temperature: float) -> float:
    """Get Z-factor for gas from Hall-Yarbrough's iterative approach.

    Parameters
    ----------
    p : float | NDArray[np.float64]
        pressure (psi)
    t : float
        temperature (Rankine)

    Returns
    -------
    zfact: float:
        z-factor

    References
    ----------
    `Hall-Yarbrough estimation <https://wiki.whitson.com/eos/eos_models/zfactor/index.html#hall-yarbrough-estimation-of-gas-z-factor>`_
    """  # noqa: E501
    t = 1 / temperature
    y = 0.001
    fdum = 1
    while np.abs(fdum) > 0.001:
        fdum = (
            -0.06125 * pressure * t * np.exp(-1.2 * (1 - t) ** 2)
            + (y + y**2 + y**3 - y**4) / (1 - y) ** 3
            - (14.76 * t - 9.76 * t**2 + 4.58 * t**3) * y**2
            + (90.7 * t - 242.2 * t**2 + 42.4 * t**3) * y ** (2.18 + 2.82 * t)
        )
        dfdy = (
            (1 + 4 * y + 4 * y**2 - 4 * y**3 + y**4) / (1 - y) ** 4
            - (29.52 * t - 19.52 * t**2 + 9.16 * t**3) * y
            + (2.18 + 2.82 * t)
            * (90.7 * t - 242.2 * t**2 + 42.4 * t**3)
            * y ** (1.18 + 2.82 * t)
        )
        y = y - fdum / dfdy
    zfact = 0.06125 * pressure * t * np.exp(-1.2 * (1 - t) ** 2) / y
    return zfact


def b_factor_DAK(
    temperature: float,
    pressure: float,
    temperature_pseudocritical: float,
    pressure_pseudocritical: float,
    temperature_standard: float = 60,
    pressure_standard: float = 14.70,
) -> float:
    """Calculate the b-factor for gas using Dranchuk and Abou-Kassem (1975).

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    pressure : float
        reservoir pressure in psia
    temperature_pseudocritical : float
        pseudocritical temperature in Fahrenheit.
    pressure_pseudocritical : float
        pseudocritical pressure in psia

    Returns
    -------
    b_g : float
        b-factor (reservoir barrels / scf)

    Examples
    -------
    >>> b_factor_DAK(400, 100, -102, 649, 60, 14.7)
    0.04317415921420302
    """
    z_factor = z_factor_DAK(
        temperature, pressure, temperature_pseudocritical, pressure_pseudocritical
    )
    b_factor = (
        z_factor
        * (temperature + 459.67)
        * pressure_standard
        / ((temperature_standard + 459.67) * pressure)
    )
    return b_factor / 5.615


def density_DAK(
    temperature: float,
    pressure: float,
    temperature_pseudocritical: float,
    pressure_pseudocritical: float,
    specific_gravity: float,
) -> float:
    """Calculate the density for gas using Dranchuk and Abou-Kassem (1975).

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    pressure : float
        reservoir pressure in psia
    temperature_pseudocritical : float
        pseudocritical temperature in Fahrenheit.
    pressure_pseudocritical : float
        pseudocritical pressure in psia
    specific_gravity : float
        specific gravity relative to air (molecular weight/ molecular weight)

    Returns
    -------
    rho_g : float
        density_gas (lb / cubic ft)

    Examples
    -------
    >>> density_DAK(400, 100, -102, 649, 0.65) # returns 0.2143
    """
    MOLECULAR_WEIGHT_AIR = 28.964
    R = 10.73159
    molecular_weight = MOLECULAR_WEIGHT_AIR * specific_gravity
    z_factor = z_factor_DAK(
        temperature, pressure, temperature_pseudocritical, pressure_pseudocritical
    )
    density_gas = pressure * molecular_weight / (z_factor * R * (temperature + 459.67))
    return density_gas


def compressibility_DAK(
    temperature: float,
    pressure: float,
    temperature_pseudocritical: float,
    pressure_pseudocritical: float,
) -> float:
    """Calculate the compressibility for gas using Dranchuk and Abou-Kassem (1975).

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    pressure : float
        reservoir pressure in psia
    temperature_pseudocritical : float
        pseudocritical temperature in Fahrenheit.
    pressure_pseudocritical : float
        pseudocritical pressure in psia

    Returns
    -------
    c_g: float
        compressibility (1 / psi)

    Examples
    ------
    >>> compressibility_DAK(400, 104.7, -102, 649)
    0.009576560643021937
    """
    A = np.array(
        [
            0.3265,
            -1.07,
            -0.5339,
            0.01569,
            -0.05165,
            0.5475,
            -0.7361,
            0.1844,
            0.1056,
            0.6134,
            0.721,
        ]
    )
    temp_reduced = (temperature + 459.67) / (temperature_pseudocritical + 459.67)
    pressure_reduced = pressure / pressure_pseudocritical
    z_factor = z_factor_DAK(
        temperature, pressure, temperature_pseudocritical, pressure_pseudocritical
    )
    rho = 0.27 * pressure_reduced / (temp_reduced * z_factor)
    dz_drho = (
        A[0]
        + A[1] / temp_reduced
        + A[2] / (temp_reduced**3)
        + A[3] / (temp_reduced**4)
        + A[4] / (temp_reduced**5)
    )
    dz_drho += 2 * (A[5] + A[6] / temp_reduced + A[7] / (temp_reduced**2)) * rho
    dz_drho += -5 * A[8] * (A[6] / temp_reduced + A[7] / (temp_reduced**2)) * rho**4
    dz_drho += (
        2 * A[9] * rho / (temp_reduced**3)
        + 2 * A[9] * A[10] * rho**3 / (temp_reduced**3)
        - 2 * A[9] * A[10] ** 2 * rho**5 / (temp_reduced**3)
    ) * math.exp(-A[10] * rho**2)
    compressibility_reduced = 1.0 / pressure_reduced - 0.27 / (
        z_factor**2 * temp_reduced
    ) * (dz_drho / (1 + rho * dz_drho / z_factor))
    compressibility = compressibility_reduced / pressure_pseudocritical
    return compressibility


def viscosity_Sutton(
    temperature: float,
    pressure: float,
    temperature_pseudocritical: float,
    pressure_pseudocritical: float,
    specific_gravity: float,
) -> float:
    """Calculate the viscosity for gas using Sutton's Fudamental PVT Calculations (2007).

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    pressure : float
        reservoir pressure in psia
    temperature_pseudocritical : float
        pseudocritical temperature in Fahrenheit.
    pressure_pseudocritical : float
        pseudocritical pressure in psia
    specific_gravity : float
        specific gravity relative to air (density/density)

    Returns
    -------
    mu_g : float
        viscosity_gas (centipoise)

    Examples
    -------
    >>> viscosity_Sutton(400, 100, -102, 649, 0.65)
    0.01652719692109309
    """
    MOLECULAR_WEIGHT_AIR = 28.964
    temp_reduced = (temperature + 459.67) / (temperature_pseudocritical + 459.67)
    molecular_weight = specific_gravity * MOLECULAR_WEIGHT_AIR
    rho = density_DAK(
        temperature,
        pressure,
        temperature_pseudocritical,
        pressure_pseudocritical,
        specific_gravity,
    )
    rho *= 16.018463 / 1e3  # convert to grams / cc
    xi = 0.949 * (
        (temperature_pseudocritical + 459.67)
        / (molecular_weight**3 * pressure_pseudocritical**4)
    ) ** (1.0 / 6.0)
    viscosity_lowpressure = (
        1e-5
        * (
            8.07 * temp_reduced**0.618
            - 3.57 * math.exp(-0.449 * temp_reduced)
            + 3.40 * math.exp(-4.058 * temp_reduced)
            + 0.18
        )
        / xi
    )
    X = 3.47 + 1588.0 / (temperature + 459.67) + 9e-4 * molecular_weight
    Y = 1.66378 - 4.679e-3 * X
    viscosity = viscosity_lowpressure * math.exp(X * rho**Y)
    return viscosity


def pseudocritical_point_Sutton(
    specific_gravity: float,
    non_hydrocarbon_properties: NDArray,
    fluid: str = "wet gas",
) -> tuple[float, float]:
    """Calculate the pseudocritical pressure and temperature from Sutton (2007).

    Parameters
    ----------
    specific_gravity : float
        specific gravity relative to air (molecular weight / molecular weight)
    non_hydrocarbon_properties: np.NDArray
        record array of non-hydrocarbon fluid properties
        **MUST HAVE H2S as second row, CO2 as third row**
    fluid : string
        whether the gas is 'dry gas' or 'wet gas'

    Returns
    -------
    pseudocritical_temp : float
        temperature_pseudocritical (F)
    pseudocritical_p : float
        pressure_pseudocritical (psia)

    Examples
    ------
    >>> non_hydrocarbon_properties = make_nonhydrocarbon_properties(0.03, 0.012, 0.018)
    >>> points_pseudocritical_Sutton(0.65, non_hydrocarbon_properties, "dry gas")
    (-102.21827232417752, 648.510797253794)
    >>> non_hydrocarbon_properties = make_nonhydrocarbon_properties(0.05, 0.01, 0.04)
    >>> points_pseudocritical_Sutton(0.8, non_hydrocarbon_properties, "wet gas")
    (-72.20351526841193, 653.2582064200534)
    """
    if fluid not in ("dry gas", "wet gas"):
        msg = f"fluid must be one of ('dry gas','wet gas'), not {fluid}"
        raise ValueError(msg)
    MOLECULAR_WEIGHT_AIR = 28.964
    fraction = non_hydrocarbon_properties["fraction"]
    molecular_weight = non_hydrocarbon_properties["molecular weight"]
    temp_critical_nonhc = non_hydrocarbon_properties["critical temperature"]
    pressure_critical_nonhc = non_hydrocarbon_properties["critical pressure"]
    fraction_hydrocarbon = 1 - non_hydrocarbon_properties["fraction"].sum()
    specific_gravity_hydrocarbon = (
        specific_gravity - sum(fraction * molecular_weight) / MOLECULAR_WEIGHT_AIR
    ) / fraction_hydrocarbon
    if fluid == "dry gas":
        temp_critical_hydrocarbon = (
            120.1
            + 429 * specific_gravity_hydrocarbon
            - 62.9 * specific_gravity_hydrocarbon**2
        )
        pressure_critical_hydrocarbon = (
            671.1
            - 14 * specific_gravity_hydrocarbon
            - 34.3 * specific_gravity_hydrocarbon**2
        )
    else:
        temp_critical_hydrocarbon = (
            164.3
            + 357.7 * specific_gravity_hydrocarbon
            - 67.7 * specific_gravity_hydrocarbon**2
        )
        pressure_critical_hydrocarbon = (
            744
            - 125.4 * specific_gravity_hydrocarbon
            + 5.9 * specific_gravity_hydrocarbon**2
        )
    temperature_star = fraction_hydrocarbon * temp_critical_hydrocarbon + sum(
        fraction * temp_critical_nonhc
    )
    pressure_star = fraction_hydrocarbon * pressure_critical_hydrocarbon + sum(
        fraction * pressure_critical_nonhc
    )
    # epsilon is the Wichert and Aziz (1970) correction
    epsilon = (
        120 * (fraction[2] + fraction[1]) ** 0.9
        - 120 * (fraction[2] + fraction[1]) ** 1.6
        + 15 * (math.sqrt(fraction[1]) - fraction[1] ** 4)
    )
    temperature_pseudocritical = temperature_star - epsilon
    pressure_pseudocritical = (
        pressure_star
        * (temperature_star - epsilon)
        / (temperature_star + fraction[1] * (1 - fraction[1]) * epsilon)
    )
    return temperature_pseudocritical - 459.67, pressure_pseudocritical


def pseudopressure_Hussainy(
    temperature: float,
    pressure: float,
    temperature_pseudocritical: float,
    pressure_pseudocritical: float,
    specific_gravity: float,
    pressure_standard: float = 14.70,
) -> float:
    """Calculate the pseudopressure for gas using Al Hussainy (1966).

    Parameters
    ----------
    temperature : float
        reservoir temperature in Fahrenheit.
    pressure : float
        reservoir pressure in psia
    temperature_pseudocritical : float
        pseudocritical temperature in Fahrenheit.
    pressure_pseudocritical : float
        pseudocritical pressure in psia
    specific_gravity : float
        specific gravity relative to air (density/density)

    Returns
    -------
    m : float
        pseudopressure (psi^2 / centipoise)

    Examples
    -------
    >>> pseudopressure_Hussainy(400, 100, -102, 649, 0.65)
    593363.7626437937
    """

    def integrand(pressure: float):
        viscosity = viscosity_Sutton(
            temperature,
            pressure,
            temperature_pseudocritical,
            pressure_pseudocritical,
            specific_gravity,
        )
        z_factor = z_factor_DAK(
            temperature, pressure, temperature_pseudocritical, pressure_pseudocritical
        )
        return 2 * pressure / (viscosity * z_factor)

    # use scipy's quadrature (wraps QUADPACK) for integration
    pseudopressure, err = quad(integrand, pressure_standard, pressure, limit=100)
    return pseudopressure



"""Forecast when bottomhole/fracface pressure is known and varying."""
from __future__ import annotations

from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy as sp
from lmfit import Minimizer, Parameters
from numpy.typing import NDArray

from bluebonnet.flow import FlowProperties, SinglePhaseReservoir


def _obj_function(
    params: Parameters,
    days: NDArray,
    production: NDArray,
    pvt_table: pd.DataFrame,
    pressure_fracface: NDArray,
) -> NDArray:
    """Calculate mismatch between scaling solution and production.

    Parameters
    ----------
    params : Parameters
        Fitting parameters
    days : NDArray
        days well is on production
    production : NDArray
        observed cumulative production over time
    pvt_table : pd.DataFrame
        pressure-volume-temperature data table
    pressure_fracface : NDArray
        pressure at the fracture face over time

    Returns
    -------
    NDArray
        mismatch between model and cumulative production
    """
    tau = params["tau"].value
    resource_in_place = params["M"].value
    pressure_initial = params["p_initial"].value
    # pressure_fracface = pressure_initial
    # print(
    #     " tau is {:7.5g}, pressure_initial is {:7.5g} and M is {:7.5g}".format(
    #         tau, pressure_initial, resource_in_place
    #     )
    # )
    t = days / tau
    flow_propertiesM = FlowProperties(pvt_table, pressure_initial)
    res_realgasM = SinglePhaseReservoir(
        80, pressure_initial, pressure_initial, flow_propertiesM
    )
    res_realgasM.simulate(t, pressure_fracface=pressure_fracface)
    recovery_factor = res_realgasM.recovery_factor()
    return resource_in_place * recovery_factor - production


def fit_production_pressure(
    prod_data: pd.DataFrame,
    pvt_table: pd.DataFrame,
    pressure_initial: float,
    filter_window_size: int | None = None,
    pressure_imax: float = 15000,
    inplace_max: float = 100000,
    filter_zero_prod_days: bool = True,
    n_iter: int = 100,
    params: Parameters | None = None,
) -> Parameters:
    """Fit cumulative production given fracface pressure.

    Parameters
    ----------
    prod_data : pd.DataFrame
        contains columns 'Days', 'Gas', and 'Pressure'
    pvt_table : pd.DataFrame
        information on equation of state, for example from build_pvt_gas
    pressure_initial : float
        guess for initial reservoir pressure
    filter_window_size : int or None
        If not None, boxcar filter size to average pressure data
    pressure_imax : float, Optional
        maximum allowed initial reservoir pressure. pvt had better include this pressure
    inplace_max : float
        Maximum allowed resource in place
    filter_zero_prod_days : bool
        Filter out days without gas production or pressure value. Also
        shortens days on production to only include productive days.
    n_iter : integer, default to 100
        number of times to iterate until stabilizes
    params : Parameters
        Initial guesses for tau, M, and p_initial. You can pass in results from previous fit.

    Returns
    -------
    params : Parameters
        Best fits for tau, M, and p_initial
    """
    if filter_zero_prod_days:
        prod_data = prod_data[
            (prod_data["Gas"] > 0) & (pd.notna(prod_data["Pressure"]))
        ][["Days", "Gas", "Pressure"]]
    else:
        prod_data = prod_data[["Days", "Gas", "Pressure"]]

    time = np.arange(0, len(prod_data["Days"]))
    pressure_fracface = np.array(prod_data["Pressure"])

    # with noisy data, sometimes a boxcar filter is beneficial
    if filter_window_size is not None:
        pressure_fracface = sp.ndimage.uniform_filter1d(
            pressure_fracface, size=filter_window_size
        )
    cumulative_prod = np.cumsum(np.array(prod_data["Gas"]))

    if params is None:
        params = Parameters()
        params.add(
            "tau", value=1000.0, min=30.0, max=time[len(time) - 1] * 2
        )  # units: days
        params.add(
            "M",
            value=cumulative_prod[-1],
            min=cumulative_prod[len(cumulative_prod) - 2],
            max=inplace_max,
        )
        params.add(
            "p_initial",
            value=pressure_initial,
            min=max(pressure_fracface),
            max=pressure_imax,
        )  # units: psi
    mini = Minimizer(
        _obj_function,
        params,
        fcn_args=(time, cumulative_prod, pvt_table, pressure_fracface),
    )
    result = mini.minimize(method="Nelder", max_nfev=n_iter)

    return result


def plot_production_comparison(
    prod_data: pd.DataFrame,
    pvt_table: pd.DataFrame,
    params: Parameters,
    filter_window_size: int | None = None,
    filter_zero_prod_days: bool = True,
    well_name: str = "Well Name",
) -> Any:
    """Compare production to match.

    Parameters
    ----------
    prod_data : pd.DataFrame
        contains columns 'Days', 'Gas', and 'Pressure'
    pvt_table : pd.DataFrame
        information on equation of state, for example from build_pvt_gas
    params : Parameters
        fit result parameters
    filter_window_size : int or None
        If not None, boxcar filter size to average pressure data
    filter_zero_prod_days : bool
        Filter out days without gas production or pressure value. Also
        shortens days on production to only include productive days.
    params : Parameters
        Best fits for tau, M, and p_initial
    well_name : str
        name to label production with

    Returns
    -------
    fig, (ax1, ax2): Any
        matplotlib figure and tuple of axes with cumulative production and
        pressure over time (scaled by tau)
    """
    if filter_zero_prod_days:
        prod_data = prod_data[
            (prod_data["Gas"] > 0) & (pd.notna(prod_data["Pressure"]))
        ][["Days", "Gas", "Pressure"]]
        time = np.arange(len(prod_data["Days"]))
    else:
        prod_data = prod_data[["Days", "Gas", "Pressure"]]
        time = prod_data["Days"]

    pressure_fracface = np.array(prod_data["Pressure"])
    #
    if filter_window_size is not None:
        pressure_fracface = sp.ndimage.uniform_filter1d(
            pressure_fracface, size=filter_window_size
        )
    #
    cumulative_prod = np.cumsum(np.array(prod_data["Gas"]))

    resource_in_place = params["M"].value
    tau = params["tau"].value
    pressure_initial = params["p_initial"].value
    # pressure_fracface = pressure_initial

    flow_propertiesM = FlowProperties(pvt_table, pressure_initial)
    res_realgasM = SinglePhaseReservoir(
        80, pressure_fracface, pressure_initial, flow_propertiesM
    )
    res_realgasM.simulate(time / tau, pressure_fracface=pressure_fracface)

    rf2M = res_realgasM.recovery_factor()
    fig, (ax1, ax2) = plt.subplots(2, 1)
    fig.set_size_inches(5, 6)
    ax1.plot(
        time / tau,
        rf2M,
        "--",
        label=f"Production; tau={tau:7.5g}, M={resource_in_place:7.5g}",
    )
    ax1.plot(time / tau, cumulative_prod / resource_in_place, label=well_name)
    ax1.legend()
    ax1.set(
        xlabel="Time",
        ylabel="Cumulative Production",
        ylim=(0, None),
        xscale="squareroot",
        xlim=(0, None),
    )

    ax2.plot(time / tau, pressure_fracface, label="Pressure (psi)")
    ax2.legend()
    ax2.set(
        xlabel="Time",
        ylabel="Pressure (psi)",
        ylim=(0, None),
        xscale="squareroot",
        xlim=(0, None),
    )

    return fig, (ax1, ax2)



"""
Fluid pressure and temperature-dependent properties.

Uses common empirical correlations to get properties. These include density,
saturation pressure, viscosity, and the pseudopressure transform.
"""
from __future__ import annotations

from bluebonnet.fluids.fluid import Fluid, build_pvt_gas, pseudopressure

__all__ = [
    "build_pvt_gas",
    "Fluid",
    "pseudopressure",
]



"""Define a suite a tests for the fluid module."""
from __future__ import annotations

import numpy as np
import pytest
from bluebonnet.fluids import Fluid


@pytest.fixture()
def fluid_instance(oil_properties):
    fluid = Fluid(
        oil_properties.temperature,
        oil_properties.api_gravity,
        oil_properties.gas_specific_gravity,
        oil_properties.solution_gor_initial,
        salinity=15,
    )
    return fluid


def test_bw(water_properties):
    real_fvf = pytest.approx(1.0312698)
    fluid_instance = Fluid(
        temperature=water_properties.temperature,
        api_gravity=0,
        gas_specific_gravity=0,
        solution_gor_initial=0,
        salinity=water_properties.salinity,
    )
    water_fvf = fluid_instance.water_FVF(np.full(2, water_properties.pressure))
    assert water_fvf[0] == real_fvf


def test_mu_w(water_properties):
    real_mu = pytest.approx(0.548030320898)
    fluid_instance = Fluid(
        temperature=water_properties.temperature,
        api_gravity=0,
        gas_specific_gravity=0,
        solution_gor_initial=0,
        salinity=water_properties.salinity,
    )
    water_mu = fluid_instance.water_viscosity(np.full(2, water_properties.pressure))
    assert water_mu[0] == real_mu


def test_bo(oil_properties):
    if np.abs(oil_properties.pressure - 3000) < 1:
        real_fvf = pytest.approx(1.381015, rel=1e-3)
    elif np.abs(oil_properties.pressure - 2000) < 1:
        real_fvf = pytest.approx(1.292999, rel=1e-3)
    else:
        msg = "haven't studied this pressure yet"
        raise ValueError(msg)
    fluid_instance = Fluid(
        temperature=oil_properties.temperature,
        api_gravity=oil_properties.api_gravity,
        gas_specific_gravity=oil_properties.gas_specific_gravity,
        solution_gor_initial=oil_properties.solution_gor_initial,
        salinity=0,
    )
    assert fluid_instance.oil_FVF(oil_properties.pressure) == real_fvf
    oil_fvf = fluid_instance.oil_FVF(np.full(2, oil_properties.pressure))
    assert oil_fvf[0] == real_fvf


def test_mu_o(oil_properties):
    if np.abs(oil_properties.pressure - 3000) < 1:
        real_mu = pytest.approx(0.511367423, rel=1e-3)
    elif np.abs(oil_properties.pressure - 2000) < 1:
        real_mu = pytest.approx(0.581137969, rel=1e-3)
    else:
        msg = "haven't studied this pressure yet"
        raise ValueError(msg)
    fluid_instance = Fluid(
        temperature=oil_properties.temperature,
        api_gravity=oil_properties.api_gravity,
        gas_specific_gravity=oil_properties.gas_specific_gravity,
        solution_gor_initial=oil_properties.solution_gor_initial,
        salinity=0,
    )
    oil_mu = fluid_instance.oil_viscosity(np.full(2, oil_properties.pressure))
    assert oil_mu[0] == real_mu


def test_bg(gas_properties):
    if gas_properties.fluid == "dry gas":
        real_fvf = pytest.approx(0.04317427, rel=1e-3)
    elif gas_properties.fluid == "wet gas":
        real_fvf = pytest.approx(0.000813928, rel=1e-3)
    else:
        ValueError("gas_properties.fluid must be one of 'wet gas' or 'dry gas'")
    fluid_instance = Fluid(
        temperature=gas_properties.temperature,
        api_gravity=10,
        gas_specific_gravity=gas_properties.specific_gravity,
        solution_gor_initial=1e3,
        salinity=0,
    )
    gas_fvf = fluid_instance.gas_FVF(
        np.full(2, gas_properties.pressure),
        gas_properties.temperature_pc,
        gas_properties.pressure_pc,
    )
    assert gas_fvf[0] == real_fvf


def test_mu_g(gas_properties):
    if gas_properties.fluid == "dry gas":
        real_mu = pytest.approx(0.016528333, rel=1e-3)
    elif gas_properties.fluid == "wet gas":
        real_mu = pytest.approx(0.023553396, rel=1e-3)
    else:
        ValueError("gas_properties.fluid must be one of 'wet gas' or 'dry gas'")
    fluid_instance = Fluid(
        temperature=gas_properties.temperature,
        api_gravity=10,
        gas_specific_gravity=gas_properties.specific_gravity,
        solution_gor_initial=1e3,
        salinity=0,
    )
    gas_mu = fluid_instance.gas_viscosity(
        np.full(2, gas_properties.pressure),
        gas_properties.temperature_pc,
        gas_properties.pressure_pc,
    )
    assert gas_mu[0] == real_mu



"""Test fitting and forecasting production."""
from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
from bluebonnet.flow import FlowProperties, IdealReservoir, SinglePhaseReservoir
from bluebonnet.forecast import (
    Bounds,
    ForecasterOnePhase,
    fit_production_pressure,
    plot_production_comparison,
)
from lmfit import Parameters

t_end = 6.0
nx = 40
nt = 2_000
pf = 500.0
pi = 5_000.0
time_scaled = np.linspace(0, np.sqrt(t_end), nt) ** 2


@pytest.fixture()
def rf_curve():
    """Return function for recovery factor from an ideal reservoir."""
    reservoir = IdealReservoir(nx, pf, pi, None)
    reservoir.simulate(time_scaled)
    reservoir.recovery_factor()
    return reservoir.recovery_factor_interpolator()


@pytest.fixture()
def pressure_varying_prod():
    """Necessary data for a pressure-varying system."""
    tau_in = 180.0
    pressure_v_time = np.full(nt, pf)
    pressure_v_time[nt // 4 : nt // 2] /= 2.0
    pressure_v_time[nt // 2 :] /= 4.0
    pvt_table = pd.read_csv("tests/data/pvt_gas_HAYNESVILLE SHALE_20.csv")
    flow_props = FlowProperties(pvt_table, pi)
    reservoir = SinglePhaseReservoir(nx, pf, pi, flow_props)
    reservoir.simulate(time_scaled, pressure_v_time)
    rf = reservoir.recovery_factor()
    prod = pd.DataFrame(
        {"Days": time_scaled * tau_in, "Gas": rf, "Pressure": pressure_v_time}
    )
    return prod, pvt_table


def test_bounds():
    """Make sure bounds uses post_init properly."""
    correct_bounds = Bounds(M=(0, 1), tau=(2, 3))
    assert correct_bounds.fit_bounds() == ((0, 2), (1, 3))
    with pytest.raises(ValueError, match="M must be two elements"):
        Bounds(M=(1, 2, 3), tau=(0, 1))
    with pytest.raises(ValueError, match="tau must be two elements"):
        Bounds(M=(1, 2), tau=(1,))
    with pytest.raises(ValueError, match="M.* must be greater"):
        Bounds(M=(1, 0), tau=(0, 1))
    with pytest.raises(ValueError, match="tau.* must be greater"):
        Bounds((0, 1), (20, 10))


def test_ForecasterOnePhase(rf_curve):
    """Test forecaster."""
    tau_test = 3.0
    M_test = 300.0
    cum_production = M_test * rf_curve(time_scaled / tau_test)

    # Fit the test production
    scaling_curve = ForecasterOnePhase(rf_curve)
    scaling_curve.fit(time_scaled, cum_production)
    M_fit = scaling_curve.M_
    tau_fit = scaling_curve.tau_
    assert M_fit == pytest.approx(M_test)
    assert tau_fit == pytest.approx(tau_test)

    # Test prediction
    cum_fit = scaling_curve.forecast_cum(time_scaled)
    assert np.allclose(cum_production, cum_fit)


def test_fit_production_pressure(pressure_varying_prod):
    """Test fitting for production and pressure."""
    prod, pvt_table = pressure_varying_prod
    result = fit_production_pressure(prod, pvt_table, pi, n_iter=4)
    assert result.params["tau"].value < 1e3, "is tau moving in the right direction?"
    assert result.params["M"].value > 1e3, "is M increasing from the initial guess?"


@pytest.mark.mpl_image_compare()
def test_fit_plot(pressure_varying_prod):
    prod, pvt_table = pressure_varying_prod
    params = Parameters()
    params.add("M", 1300)
    params.add("tau", 420)
    params.add("p_initial", pi)
    fig, _ = plot_production_comparison(
        prod,
        pvt_table,
        params,
        filter_window_size=1,
        filter_zero_prod_days=True,
    )
    return fig



"""reservoir: scaling solutions for hydrofractured wells.

This module calculates the scaling curves, using flow properties and finite
difference methods.
"""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass

import numpy as np
import numpy.typing as npt
from numpy import ndarray
from scipy import integrate, interpolate, sparse

from bluebonnet.flow.flowproperties import FlowProperties

_ATOL = 1e-12


@dataclass
class IdealReservoir:
    """
    Class for building scaling solutions of production from hydrofractured wells.

    This maintains simulation parameters and fluid properties.

    Parameters
    ----------
    nx : int
        number of spatial nodes
    pressure_fracface : float | NDArray
        drawdown pressure at x=0 (psi)
    pressure_initial : float
        reservoir pressure before production (psi)
    fluid : FlowProperties
        reservoir fluid PVT/flow properties
    """

    nx: int
    """number of spatial nodes"""
    pressure_fracface: float | npt.NDArray
    """drawdown pressure at :math:`x=0` (psi)"""
    pressure_initial: float
    """reservoir pressure before production (psi)"""
    fluid: FlowProperties
    """reservoir fluid PVT/flow properties"""

    def __post_init__(self):
        """Last initialization steps."""

    def simulate(self, time: ndarray):
        """
        Calculate simulation pressure over time.

        Parameters
        ----------
        time : ndarray
            times to solve for pressure
        """
        self.time = time
        x = np.linspace(0, 1, self.nx)
        dx_squared = (x[1] - x[0]) ** 2
        pseudopressure = np.empty((len(time), self.nx))
        pseudopressure[0, :] = 1.0
        for i in range(time.shape[0] - 1):
            b = pseudopressure[i]
            mesh_ratio = (time[i + 1] - time[i]) / dx_squared
            alpha_scaled = self.alpha_scaled(b)
            kt_h2 = mesh_ratio * alpha_scaled
            a_matrix = _build_matrix(kt_h2)
            pseudopressure[i + 1], _ = sparse.linalg.bicgstab(a_matrix, b, atol=_ATOL)
        self.pseudopressure = pseudopressure

    def recovery_factor(self, time: ndarray | None = None, density=False) -> ndarray:
        """Calculate recovery factor over time.

        If time has is not specified, requires that `simulate` has been run

        Parameters
        ----------
        time : ndarray, Optional
            times to calculate recovery factor at

        Returns
        -------
        recovery : ndarray
            recovery factor over time
        """
        if time is None:
            try:
                time = self.time
            except AttributeError as e:
                msg = "Need to run simulate before calculating recovery factor"
                raise RuntimeError(msg) from e
        h_inv = self.nx - 1.0
        if density:
            pseudopressure_to_mass = interpolate.interp1d(
                self.fluid.pvt_props["m-scaled"],
                self.fluid.pvt_props["density"],
                fill_value="extrapolate",
            )
            mass = pseudopressure_to_mass(self.pseudopressure)
            mass_over_time = np.sum(mass, 1)
            cumulative = 1.0 - mass_over_time / mass_over_time[0]
        else:
            pp = self.pseudopressure[:, :3]
            rate = (-pp[:, 2] + 4 * pp[:, 1] - 3 * pp[:, 0]) * h_inv * 0.5  # dp_dx
            cumulative = integrate.cumulative_trapezoid(rate, self.time, initial=0)
        self.recovery = cumulative * self.fvf_scale()
        return self.recovery

    def recovery_factor_interpolator(self) -> Callable:
        """Generate a function to get recovery factor from time.

        Requires that `recovery_factor` has been run

        Returns
        -------
        scipy interpolator object that takes in time and spits out recovery factor
        """
        try:
            time = self.time
        except AttributeError as e:
            msg = "Need to run simulate"
            raise RuntimeError(msg) from e
        try:
            recovery = self.recovery
        except AttributeError:
            recovery = self.recovery_factor(time)

        interpolator = interpolate.interp1d(
            time, recovery, bounds_error=False, fill_value=(0, recovery[-1])
        )
        return interpolator

    def alpha_scaled(self, pseudopressure: ndarray) -> ndarray:
        """Calculate scaled diffusivity."""
        return np.ones_like(pseudopressure)

    def fvf_scale(self) -> float:
        """Scaling for formation volume factor.

        Returns
        -------
        FVF : float
        """
        return 1 - self.pressure_fracface / self.pressure_initial


class SinglePhaseReservoir(IdealReservoir):
    """Single-phase real fluid reservoir."""

    def alpha_scaled(self, pseudopressure: ndarray) -> ndarray:
        """Calculate scaled diffusivity."""
        alpha = self.fluid.alpha
        return alpha(pseudopressure) / alpha(self.fluid.m_i)

    def fvf_scale(self) -> float:
        """Scaling for formation volume factor.

        Returns
        -------
        FVF : float
        """
        return 1

    def simulate(
        self, time: ndarray[float], pressure_fracface: ndarray[float] | None = None
    ):
        """Calculate simulation pressure over time.

        Args
        ----
        time : ndarray
            times to solve for pressure
        pressure_fracface : Iterable[float] | None, optional
            pressure at frac-face over time. Defaults to None, which is no change

        Raises
        ------
        ValueError: wrong length changing pressure at frac-face
        """
        self.time = time
        dx_squared = (1 / self.nx) ** 2
        pseudopressure = np.empty((len(time), self.nx))
        if pressure_fracface is None:
            pressure_fracface = np.full(len(time), self.pressure_fracface)
        else:
            if len(pressure_fracface) != len(time):
                msg = (
                    "Pressure time series does not match time variable:"
                    f" {len(pressure_fracface)} versus {len(time)}"
                )
                raise ValueError(msg)
            self.pressure_fracface = pressure_fracface
        m_i = self.fluid.m_i
        m_f = self.fluid.m_scaled_func(pressure_fracface)
        pseudopressure_initial = np.full(self.nx, m_i)
        pseudopressure_initial[0] = m_f[0]
        pseudopressure[0, :] = pseudopressure_initial

        for i in range(len(time) - 1):
            mesh_ratio = (time[i + 1] - time[i]) / dx_squared
            b = np.minimum(pseudopressure[i].copy(), m_i)
            # Enforce the boundary condition at x=0
            b[0] = m_f[i] + self.alpha_scaled(m_f[i]) * m_f[i] * mesh_ratio
            try:
                alpha_scaled = self.alpha_scaled(b)
            except ValueError as e:
                msg = f"scaling failed where m_initial={m_i}  m_fracface={m_f}"
                raise ValueError(msg) from e
            kt_h2 = mesh_ratio * alpha_scaled
            a_matrix = _build_matrix(kt_h2)
            pseudopressure[i + 1], _ = sparse.linalg.bicgstab(a_matrix, b, atol=_ATOL)
        self.pseudopressure = pseudopressure


@dataclass
class TwoPhaseReservoir(SinglePhaseReservoir):
    """Oil-gas reservoir simulation.

    References
    ----------
    Ruiz Maraggi, L.M., Lake, L.W. and Walsh, M.P., 2020. "A Two-Phase Non-Linear One-
    Dimensional Flow Model for Reserves Estimation in Tight Oil and Gas
    Condensate Reservoirs Using Scaling Principles." In SPE Latin American and
    Caribbean Petroleum Engineering Conference. OnePetro.
    https://doi.org/10.2118/199032-MS
    """

    Sw_init: float

    def simulate(self, time: ndarray):
        """Calculate simulation pressure over time.

        Parameters
        ----------
        time : ndarray
            times to solve for pressure
        """
        super().simulate(time)


@dataclass
class MultiPhaseReservoir(SinglePhaseReservoir):
    """Reservoir with three phases: oil, gas, and water all flowing.

    Parameters
    ----------
    nx : int
        number of spatial nodes
    pressure_fracface : float | NDArray
        drawdown pressure at x=0 (psi)
    pressure_initial : float
        reservoir pressure before production (psi)
    fluid : FlowProperties
        reservoir fluid PVT/flow properties
    So_init : float
        oil saturation at the beginning
    Sw_init : float
        water saturation
    Sg_init : float
        gas saturation
    """

    So_init: float
    """Initial oil saturation."""
    Sw_init: float
    """Initial water saturation."""
    Sg_init: float
    """Initial gas saturation."""

    def simulate(self, time: ndarray):
        """Calculate simulation pressure over time.

        Parameters
        ----------
        time : ndarray
            times to solve for pressure
        """
        raise NotImplementedError  # TODO: saturation changes

        x = np.linspace(0, 1, self.nx)
        dx_squared = (x[1] - x[0]) ** 2
        pseudopressure = np.empty((len(time), self.nx))
        pseudopressure[0, :] = 1
        pseudopressure[0, 0] = 0
        sat_names = ("So", "Sg", "Sw")
        saturations = np.empty(
            (len(time), self.nx),
            dtype=[(s, np.float64) for s in sat_names],
        )
        for i in range(time.shape[0] - 1):
            b = pseudopressure[i]
            sat = saturations[i]
            dt = time[i + 1] - time[i]
            mesh_ratio = dt / dx_squared
            alpha_scaled = self.alpha_scaled(b, sat)
            kt_h2 = mesh_ratio * alpha_scaled
            a_matrix = _build_matrix(kt_h2)
            pseudopressure[i + 1], info = sparse.linalg.bicgstab(
                a_matrix, b, atol=_ATOL
            )
            saturations[i + 1] = self._step_saturation(sat, b, pseudopressure[i + 1])
        self.time = time
        self.pseudopressure = pseudopressure

    def alpha_scaled(
        self,
        pseudopressure: ndarray,
        saturation: ndarray,
    ) -> ndarray:
        """
        Calculate scaled diffusivity given pseudopressure and saturations.

        Parameters
        ----------
        pseudopressure : ndarray
            scaled pseudopressure
        saturation : ndarray
            record array with So, Sg, Sw records

        Returns
        -------
        alpha : ndarray
            scaled diffusivity
        """
        alpha = self.fluid.alpha
        s = saturation
        return alpha(pseudopressure, s["So"], s["Sg"], s["Sw"]) / alpha(1)

    def _step_saturation(
        self,
        saturation: ndarray,  # noqa: ARG002
        ppressure_old: ndarray,  # noqa: ARG002
        ppressure_new: ndarray,  # noqa: ARG002
    ) -> ndarray:
        """Calculate new saturation.

        Args
        ----
        saturation : ndarray
            old saturation
        ppressure_old : ndarray
            old pressure
        ppressure_new : ndarray
            new pressure

        Returns
        -------
        new saturation: ndarray
        """
        return NotImplementedError


def _build_matrix(kt_h2: ndarray) -> sparse.spmatrix:
    r"""
    Set up A matrix for timestepping.

    Follows :math: `A x = b` -> :math: `x = A \ b`

    Parameters
    ----------
    kt_h2: ndarray
        diffusivity * dt / dx^2

    Returns
    -------
    a_matrix: sp.sparse.matrix
        The A matrix
    """
    diagonal_long = 1.0 + 2 * kt_h2
    # diagonal_long[0] = -1.0
    # diagonal_low = np.concatenate([[0], -kt_h2[2:-1], [-2 * kt_h2[-1]]])
    # diagonal_upper = np.concatenate([[0, -kt_h2[1]], -kt_h2[2:-1]])
    diagonal_long[-1] = 1.0 + kt_h2[-1]
    diagonal_low = -kt_h2[1:]
    diagonal_upper = -kt_h2[0:-1]
    a_matrix = sparse.diags(
        [diagonal_low, diagonal_long, diagonal_upper], [-1, 0, 1], format="csr"
    )
    return a_matrix



"""Define a suite a tests for the oil module."""
from __future__ import annotations

import math

import pytest
from bluebonnet.fluids import oil

TEMPERATURE_STANDARD = 60.0
PRESSURE_STANDARD = 14.70


def test_bubblepoint_pressure_Standing(oil_properties):
    """Test Standing bubblepoint pressure calculation."""
    (
        temperature,
        pressure,
        api_gravity,
        gas_specific_gravity,
        solution_gor_initial,
        fluid,
    ) = oil_properties
    pressure_bubblepoint_real = pytest.approx(2603.1217021875277)
    pressure_bubblepoint = oil.pressure_bubblepoint_Standing(
        temperature, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    assert pressure_bubblepoint == pressure_bubblepoint_real


def test_solution_gor_Standing(oil_properties):
    """Test Standing calculation of solution gas:oil raio."""
    (
        temperature,
        pressure,
        api_gravity,
        gas_specific_gravity,
        solution_gor_initial,
        fluid,
    ) = oil_properties
    if fluid == "black oil" and math.fabs(pressure - 3000) < 1:
        gor_real = pytest.approx(650)
    elif fluid == "black oil" and math.fabs(pressure - 2000) < 1:
        gor_real = pytest.approx(474.8228813376839)
    else:
        raise NotImplementedError
    gor_standing = oil.solution_gor_Standing(
        temperature, pressure, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    assert gor_standing == gor_real


def test_dgor_dpressure_Standing(oil_properties):
    """Test whether Standing gets change in GOR over pressure right."""
    (
        temperature,
        pressure,
        api_gravity,
        gas_specific_gravity,
        solution_gor_initial,
        fluid,
    ) = oil_properties
    if fluid == "black oil" and math.fabs(pressure - 3000) < 1:
        dgor_real = pytest.approx(0)
    elif fluid == "black oil" and math.fabs(pressure - 2000) < 1:
        dgor_real = pytest.approx(0.2824395998221715)
    else:
        raise NotImplementedError
    dgor_standing = oil.dgor_dpressure_Standing(
        temperature, pressure, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    assert dgor_standing == dgor_real


def test_B_o_bubblepoint_Standing(oil_properties):
    """Test Standing bubblepoint formation volume factor."""
    (
        temperature,
        pressure,
        api_gravity,
        gas_specific_gravity,
        solution_gor_initial,
        fluid,
    ) = oil_properties
    B_o_bubblepoint_real = pytest.approx(1.3860514623492897)
    B_o_Standing = oil.b_o_bubblepoint_Standing(
        temperature, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    assert B_o_bubblepoint_real == B_o_Standing


def test_oil_compressibility_undersat_Standing(oil_properties):
    """Test Standing oil compressibility for undersaturated oil."""
    (
        temperature,
        pressure,
        api_gravity,
        gas_specific_gravity,
        solution_gor_initial,
        fluid,
    ) = oil_properties

    # overwrite pressure to be above saturation pressure
    pressure = 3000.0
    C_o_real = pytest.approx(1.4719851e-05)
    C_o_Standing = oil.oil_compressibility_undersat_Standing(
        temperature, pressure, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    assert C_o_Standing == C_o_real


def test_oil_compressibility_undersat_Spivey(oil_properties):
    """Test Spivey's oil compressibility for undersaturated oil."""
    (
        temperature,
        pressure,
        api_gravity,
        gas_specific_gravity,
        solution_gor_initial,
        _,
    ) = oil_properties

    # overwrite pressure to be above saturation pressure
    if pressure < 2001.0:
        pytest.skip("only makes sense if undersaturated")
    C_o_real = pytest.approx(9.1723089e-6)
    C_o_Spivey = oil.oil_compressibility_undersat_Spivey(
        temperature, pressure, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    assert C_o_Spivey == C_o_real


def test_B_o_Standing(oil_properties):
    """Test Standing's formation volume factor (all saturations) for oil."""
    (
        temperature,
        pressure,
        api_gravity,
        gas_specific_gravity,
        solution_gor_initial,
        fluid,
    ) = oil_properties
    if fluid == "black oil" and math.fabs(pressure - 3000) < 1:
        B_o_real = pytest.approx(1.38101, rel=1e-3)
    elif fluid == "black oil" and math.fabs(pressure - 2000) < 1:
        B_o_real = pytest.approx(1.2929990, rel=1e-3)
    else:
        raise NotImplementedError
    B_o_Standing = oil.b_o_Standing(
        temperature, pressure, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    assert B_o_Standing == B_o_real


def test_oil_compressibility_Standing(oil_properties):
    """Test Standing's oil compressibility (all saturations)."""
    (
        temperature,
        pressure,
        api_gravity,
        gas_specific_gravity,
        solution_gor_initial,
        fluid,
    ) = oil_properties
    pseudocritical_temperature = -102.21827232417752
    pseudocritical_pressure = 653.258206420053
    if fluid == "black oil" and math.fabs(pressure - 3000) < 1:
        C_o_real = pytest.approx(9.172308e-06, rel=1e-3)
    elif fluid == "black oil" and math.fabs(pressure - 2000) < 1:
        C_o_real = pytest.approx(2.11633e-4, rel=1e-3)
    else:
        raise NotImplementedError
    C_o_Standing = oil.oil_compressibility_Standing(
        temperature,
        pressure,
        api_gravity,
        gas_specific_gravity,
        solution_gor_initial,
        pseudocritical_temperature,
        pseudocritical_pressure,
    )
    assert C_o_Standing == C_o_real


def test_oil_density_Standing(oil_properties):
    """Test Standing's density (depends on solution GOR, b_o)."""
    (
        temperature,
        pressure,
        api_gravity,
        gas_specific_gravity,
        solution_gor_initial,
        fluid,
    ) = oil_properties
    if fluid == "black oil" and math.fabs(pressure - 3000) < 1:
        density_real = pytest.approx(43.50215, rel=1e-3)
    elif fluid == "black oil" and math.fabs(pressure - 2000) < 1:
        density_real = pytest.approx(45, rel=0.2)  # for now until it works
    else:
        raise NotImplementedError

    density_Standing = oil.density_Standing(
        temperature, pressure, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    assert density_Standing == density_real


def test_viscosity_beggs_robinson(oil_properties):
    """Test Beggs-Robinson viscosity (depends on solution GOR, bubble point)."""
    (
        temperature,
        pressure,
        api_gravity,
        gas_specific_gravity,
        solution_gor_initial,
        fluid,
    ) = oil_properties
    if fluid == "black oil" and math.fabs(pressure - 3000) < 1:
        viscosity_real = pytest.approx(0.5113674, rel=1e-3)
    elif fluid == "black oil" and math.fabs(pressure - 2000) < 1:
        viscosity_real = pytest.approx(0.5811379, rel=1e-3)
    else:
        msg = "pressure tests only cover 3000 and 2000"
        raise ValueError(msg)
    viscosity_br = oil.viscosity_beggs_robinson(
        temperature, pressure, api_gravity, gas_specific_gravity, solution_gor_initial
    )
    assert viscosity_br == viscosity_real



"""Fit and forecast production.

For fitting one-phase or pseudo-one-phase production, use ForecasterOnePhase.

To fit production when fracface pressure is known, use fit_production_pressure.
"""

# read version from installed package
from __future__ import annotations

from importlib.metadata import version

__version__ = version("bluebonnet")

from bluebonnet.forecast.forecast import Bounds, ForecasterOnePhase
from bluebonnet.forecast.forecast_pressure import (
    fit_production_pressure,
    plot_production_comparison,
)

__all__ = [
    "Bounds",
    "ForecasterOnePhase",
    "fit_production_pressure",
    "plot_production_comparison",
]



"""Nox sessions for linting, docs, and testing."""
from __future__ import annotations

import os
import shutil
from pathlib import Path

import nox

DIR = Path(__file__).parent.resolve()

nox.options.sessions = ["lint", "tests"]


@nox.session
def lint(session: nox.Session) -> None:
    """Run the linter.

    Includes all the pre-commit checks on all the files.
    """
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@nox.session
def tests(session: nox.Session) -> None:
    """Run the unit and regular tests."""
    session.install(".[test]")
    session.run("pytest", *session.posargs)


@nox.session
def docs(session: nox.Session) -> None:
    """Build the docs. Pass "serve" to serve."""
    session.install(".[docs]")
    session.chdir("docs")
    session.run("sphinx-build", "-M", "html", ".", "_build")

    if session.posargs:
        if "serve" in session.posargs:
            print("Launching docs at http://localhost:8000/ - use Ctrl-C to quit")
            session.run("python", "-m", "http.server", "8000", "-d", "_build/html")
        else:
            session.warn("Unsupported argument to docs")


@nox.session
def build(session: nox.Session) -> None:
    """Build an SDist and a wheel."""
    build_p = DIR.joinpath("build")
    if build_p.exists():
        shutil.rmtree(build_p)

    session.install("build")
    session.run("python", "-m", "build")


@nox.session
def paper(session: nox.Sesson) -> None:
    """Build the JOSS paper draft."""
    paper_dir = DIR.joinpath("paper")
    session.run(
        "docker",
        "run",
        "--rm",
        "--volume",
        f"{paper_dir}:/data",
        "--user",
        f"{os.getuid()}:{os.getgid()}",
        "--env",
        "JOURNAL=joss",
        "openjournals/inara",
        external=True,
    )



"""Define a suite a tests for the gas module."""
from __future__ import annotations

import pytest
from bluebonnet.fluids import water

TEMPERATURE_STANDARD = 60.0
PRESSURE_STANDARD = 14.70


def test_b_water_McCain(water_properties):
    temperature = water_properties.temperature
    pressure = water_properties.pressure
    b_water_true = pytest.approx(1.0312698)
    assert water.b_water_McCain(temperature, pressure) == b_water_true


def test_b_water_McCain_dp(water_properties):
    temperature = water_properties.temperature
    pressure = water_properties.pressure
    b_water_dp_true = pytest.approx(-2.9382699e-06)
    assert water.b_water_McCain_dp(temperature, pressure) == b_water_dp_true


def test_compressibility_McCain(water_properties):
    temperature = water_properties.temperature
    pressure = water_properties.pressure
    salinity = water_properties.salinity
    co_water_true = pytest.approx(3.0860375940019587e-06)
    assert (
        water.compressibility_water_McCain(temperature, pressure, salinity)
        == co_water_true
    )


def test_density_McCain(water_properties):
    density_water_true = pytest.approx(67.205706)
    assert (
        water.density_water_McCain(
            water_properties.temperature,
            water_properties.pressure,
            water_properties.salinity,
        )
        == density_water_true
    )


def test_viscosity_McCain(water_properties):
    viscosity_water_true = pytest.approx(0.548030320898)
    assert (
        water.viscosity_water_McCain(
            water_properties.temperature,
            water_properties.pressure,
            water_properties.salinity,
        )
        == viscosity_water_true
    )
